# mypy: disable-error-code="call-overload, misc, override"

from typing import Literal as _Literal

from pydantic import Field as _Field

from pyswap.components.tables import INITSOILTEMP, SOILTEXTURES
from pyswap.core.basemodel import PySWAPBaseModel as _PySWAPBaseModel
from pyswap.core.fields import (
    Decimal2f as _Decimal2f,
    String as _String,
    Table as _Table,
)
from pyswap.core.valueranges import YEARRANGE as _YEARRANGE
from pyswap.utils.mixins import (
    SerializableMixin as _SerializableMixin,
    YAMLValidatorMixin as _YAMLValidatorMixin,
)

__all__ = ["INITSOILTEMP", "SOILTEXTURES", "HeatFlow", "SoluteTransport"]


class HeatFlow(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Heat flow settings for SWAP simulation.

    Attributes:
        swhea (Literal[0, 1]): Switch for heat flow.
        swcalt (Optional[Literal[1, 2]]):
            * 1 - analytical method
            * 2 - numerical method
        tampli (Optional[Decimal2f]): Amplitude of annual temperature wave at soil surface [0..50 oC, R]
        tmean (Optional[Decimal2f]): Mean annual temperature at soil surface [-10..30 oC, R]
        timref (Optional[Decimal2f]): Time at which the sinus temperature wave reaches its top [0..366.0 d, R]
        ddamp (Optional[Decimal2f]): Damping depth of soil temperature wave [1..500 cm, R]
        swtopbhea (Optional[Literal[1, 2]]): Define top boundary condition
            * 1 - use air temperature of meteo input file as top boundary
            * 2 - use measured top soil temperature as top boundary
        tsoilfile (Optional[str]): Name of input file with soil surface temperatures without extension .TSS
        swbotbhea (Optional[Literal[1, 2]]): Define bottom boundary condition
            * 1 - no heat flux
            * 2 - prescribe bottom temperature
        soiltextures (Optional[Table]): For each physical soil layer the soil texture (g/g mineral parts) and the organic matter content (g/g dry soil)
        initsoiltemp (Optional[Table]): Initial temperature TSOIL [-50..50 oC, R] as function of soil depth ZH [-100000..0 cm, R]
        bbctsoil (Optional[Table]): Bottom boundary temperature TBOT [-50..50 oC, R] as function of date DATET [date]
    """

    swhea: _Literal[0, 1] | None = None
    swcalt: _Literal[1, 2] | None = None
    tampli: _Decimal2f | None = _Field(None, ge=0, le=50)
    tmean: _Decimal2f | None = _Field(None, ge=-10, le=30)
    timref: _Decimal2f | None = _Field(None, **_YEARRANGE)
    ddamp: _Decimal2f | None = _Field(None, ge=1, le=500)
    swtopbhea: _Literal[1, 2] | None = None
    tsoilfile: _String | None = None
    swbotbhea: _Literal[1, 2] | None = None
    soiltextures: _Table | None = None
    initsoiltemp: _Table | None = None
    bbctsoil: _Table | None = None


class SoluteTransport(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Solute transport settings.

    Attributes:

        swsolu (Literal[0, 1]): Switch for simulation of solute transport.
        cpre (Optional[Decimal2f]): Solute concentration in precipitation [0..100 mg/cm3].
        cdrain (Optional[Decimal2f]): Solute concentration in surface water [0..100 mg/cm3].
        swbotbc (Optional[Literal[0, 1, 2]]): Switch for groundwater concentration in case of upward flow (seepage).
        cseep (Optional[Decimal2f]): Solute concentration in surface water [0..100 mg/cm3].
        ddif (Optional[Decimal2f]): Molecular diffusion coefficient [0..10 cm2/d].
        tscf (Optional[Decimal2f]): Relative uptake of solutes by roots [0..10].
        swsp (Optional[Literal[0, 1]]): Switch, consider solute adsorption.
        frexp (Optional[Decimal2f]): Freundlich exponent [0..10].
        cref (Optional[Decimal2f]): Reference solute concentration for adsorption [0..1000 mg/cm3].
        swdc (Optional[Literal[0, 1]]): Switch, consider solute decomposition.
        gampar (Optional[Decimal2f]): Factor reduction decomposition due to temperature [0..0.5 /C].
        rtheta (Optional[Decimal2f]): Minimum water content for potential decomposition [0..0.4 cm3/cm3].
        bexp (Optional[Decimal2f]): Exponent in reduction decomposition due to dryness [0..2].
        swbr (Optional[Literal[0, 1]]): Switch, consider mixed reservoir of saturated zone.
        daquif (Optional[Decimal2f]): Thickness saturated part of aquifer [0..10000 cm].
        poros (Optional[Decimal2f]): Porosity of aquifer [0..0.6].
        kfsat (Optional[Decimal2f]): Linear adsorption coefficient in aquifer [0..100 cm3/mg].
        decsat (Optional[Decimal2f]): Decomposition rate in aquifer [0..10 /d].
        cdraini (Optional[Decimal2f]): Initial solute concentration in groundwater [0..100 mg/cm3].
        cseeparrtb (Optional[Table]): Table for groundwater concentration as function of time.
        inissoil (Optional[Table]): Table for initial solute concentration as function of soil depth.
        miscellaneous (Optional[Table]): Table for miscellaneous parameters as function of soil depth.
    """

    swsolu: _Literal[0, 1] | None = None
    cpre: _Decimal2f | None = _Field(None, ge=0, le=100)
    cdrain: _Decimal2f | None = _Field(None, ge=0, le=100)
    swbotbc: _Literal[0, 1, 2] | None = None
    cseep: _Decimal2f | None = _Field(None, ge=0, le=100)
    ddif: _Decimal2f | None = _Field(None, ge=0, le=10)
    tscf: _Decimal2f | None = _Field(None, ge=0, le=10)
    swsp: _Literal[0, 1] | None = None
    frexp: _Decimal2f | None = _Field(None, ge=0, le=10)
    cref: _Decimal2f | None = _Field(None, ge=0, le=1000)
    swdc: _Literal[0, 1] | None = None
    gampar: _Decimal2f | None = _Field(None, ge=0, le=0.5)
    rtheta: _Decimal2f | None = _Field(None, ge=0, le=0.4)
    bexp: _Decimal2f | None = _Field(None, ge=0, le=2)
    swbr: _Literal[0, 1] | None = None
    daquif: _Decimal2f | None = _Field(None, ge=0, le=10000)
    poros: _Decimal2f | None = _Field(None, ge=0, le=0.6)
    kfsat: _Decimal2f | None = _Field(None, ge=0, le=100)
    decsat: _Decimal2f | None = _Field(None, ge=0, le=10)
    cdraini: _Decimal2f | None = _Field(None, ge=0, le=100)
    cseeparrtb: _Table | None = None
    inissoil: _Table | None = None
    misc: _Table | None = None



"""Interact with the filesystem

All functions that interact with the filesystem are located in this subpackage.

Modules:
    io_ascii: Functions to interact with ASCII files.
    io_yaml: Functions to interact with YAML files.
    classic_swap: Functions to load with classic SWAP input files.
"""



# mypy: disable-error-code="attr-defined"

import pyswap as psp
from pyswap import testcase


def _make_grassgrowth():
    """Loading the grassgrowth model from ascii files."""
    meta = psp.components.Metadata(
        author="John Doe",
        institution="University of Somewhere",
        email="john.doe@somewhere.com",
        project="pySWAP test - hupselbrook",
        swap_ver="4.2",
    )

    ml = psp.load_swp(testcase.get_path("grassgrowth", "swp"), meta)
    ml.bottomboundary = psp.load_bbc(
        testcase.get_path("grassgrowth", "bbc"), ml.bottomboundary
    )

    ml.crop.cropfiles = {
        "grassd": psp.load_crp(
            testcase.get_path("grassgrowth", "grassd"), crptype="grass", name="grassd"
        )
    }

    ml.lateraldrainage.drafile = psp.load_dra(testcase.get_path("grassgrowth", "dra"))

    ml.meteorology.metfile = psp.components.meteorology.metfile_from_csv(
        "260.met", testcase.get_path("grassgrowth", "met")
    )
    return ml



# Version information
import importlib.metadata
import logging

from pyswap import components, db, gis, testcase

# # This will be better if called as `from pyswap.db import HDF5`, makes it clear
# # where it is coming from
from pyswap.core import io, plot
from pyswap.log import set_log_level, setup_logging
from pyswap.model.model import Model, run_parallel
from pyswap.utils.executables import check_swap, get_swap, show_info
from pyswap.utils.loaders import load_bbc, load_crp, load_dra, load_swp

__version__ = importlib.metadata.version("pyswap")

__all__ = [
    "Model",
    "check_swap",
    "components",
    "db",
    "get_swap",
    "gis",
    "io",
    "load_bbc",
    "load_crp",
    "load_dra",
    "load_swp",
    "plot",
    "run_parallel",
    "set_log_level",
    "setup_logging",
    "show_info",
    "testcase",
]


logging.getLogger("pyswap").addHandler(logging.NullHandler())



"""Loading datasets for testcases."""

import platform
from pathlib import Path

import pandas as pd

IS_WINDOWS = platform.system() == "Windows"

BASE_PATH: Path = Path(__file__).parent.joinpath("./data")

RESOURCES: dict[str, dict[str, Path]] = {
    "hupselbrook": {
        "met": Path(BASE_PATH, "1-hupselbrook/283.csv"),
        "swp": Path(BASE_PATH, "1-hupselbrook/swap.swp"),
        "dra": Path(BASE_PATH, "1-hupselbrook/swap.dra"),
        "maizes": Path(BASE_PATH, "1-hupselbrook/maizes.crp"),
        "maize_tables_yaml": Path(BASE_PATH, "1-hupselbrook/maize_swap.yaml"),
        "potatod": Path(BASE_PATH, "1-hupselbrook/potatod.crp"),
        "potato_tables_yaml": Path(BASE_PATH, "1-hupselbrook/potato_swap.yaml"),
        "grassd": Path(BASE_PATH, "1-hupselbrook/grassd.crp"),
        "grass_tables_yaml": Path(BASE_PATH, "1-hupselbrook/grass_swap.yaml"),
    },
    "grassgrowth": {
        "met": Path(BASE_PATH, "2-grassgrowth/260.csv"),
        "grassd": Path(BASE_PATH, "2-grassgrowth/grassd.crp"),
        "bbc": Path(BASE_PATH, "2-grassgrowth/swap.bbc"),
        "dra": Path(BASE_PATH, "2-grassgrowth/swap.dra"),
        "swp": Path(BASE_PATH, "2-grassgrowth/swap.swp"),
        "gwlevels": Path(BASE_PATH, "2-grassgrowth/gwlevels.csv"),
        "observations": Path(BASE_PATH, "2-grassgrowth/yield.csv"),
    },
    "macroporeflow": {"met": Path(BASE_PATH, "3-macropore-flow/andelst_meteo.met")},
    "oxygenstress": {"met": Path(BASE_PATH, "4-oxygenstress/260.csv")},
}


def view_resources(testcase: str) -> dict:
    return RESOURCES[testcase]


def get_path(testcase: str, fextension: str) -> Path:
    return RESOURCES[testcase].get(fextension)


def load_met(testcase: str, **kwargs) -> pd.DataFrame:
    path = RESOURCES[testcase].get("met")
    return pd.read_csv(path, **kwargs)


def load_observations(testcase: str, **kwargs) -> pd.DataFrame:
    path = RESOURCES[testcase].get("observations")
    return pd.read_csv(path, **kwargs)



# ruff: noqa: S301, TRY300
# mypy: disable-error-code="operator, index"
# operator and index errors were showing up in the _load_pickled method. This
# should be observed and fixed in the future if needed.

# Perhaps in the next versions, we can implement methods to handle pickes more
# safely. For now, it's not a high priority.
# Error TRY300 also seemed a bit too strict for this case.
"""Interacting with HDF5 files.

HDF5 database is meant to store pySWAP models and results in a structured way.
The class allows users to save and load multiple models and results to a single
.h5 file.

Classes:
    HDF5: Class for interacting with HDF5 files.
"""

import logging
import pickle
from typing import TYPE_CHECKING, Literal, Union

import h5py
import numpy as np
from pydantic import BaseModel, Field, computed_field

if TYPE_CHECKING:
    from pyswap.model.model import Model
    from pyswap.model.result import Result

logger = logging.getLogger(__name__)


class HDF5(BaseModel):
    """Class for interacting with HDF5 files.

    The pySWAP models and results are stored in the HDF5 file as pickled
    objects. These are complete representations of the models and results
    that can be loaded back into memory. They also have some metadata attached
    to them which are stored as atrributes in the HDF5 file. Users can view the
    models in an HDF5 file viewer.

    Attributes:
        filename (str): Path to the HDF5 file.
        models (dict): Dictionary containing the loaded models.
    """

    filename: str
    models: dict | None = Field(default_factory=dict)

    @computed_field(return_type=dict)
    def list_projects(self):
        """List all the projects in the HDF5 file."""
        with h5py.File(self.filename, "r") as f:
            projects = list(f.keys())
        return projects

    @computed_field(return_type=dict)
    def list_models(self):
        """List all the models in the HDF5 file."""
        with h5py.File(self.filename, "r") as f:
            # Use the visititems method to traverse the file structure
            models = {project: list(f[project].keys()) for project in f}
        return models

    @staticmethod
    def _get_or_create_group(f, group_name):
        """Get a group from an HDF5 file or create it if it does not exist."""
        return f.require_group(group_name)

    def save_model(
        self,
        model: "Model",
        result: Union["Result", None] = None,
        overwrite_datasets: bool = False,
        overwrite_project: bool = False,
        mode: Literal["python", "json", "yaml", "plain"] = "python",
    ):
        """
        Save a model and its results to an HDF5 file.

        Parameters:
            model (Model): The model to be saved.
            result (Result, optional): The result to be saved.
            overwrite_datasets (bool): If True, overwrite datasets if they exist.
            overwrite_project (bool): If True, overwrite the project if it exists.
            mode (Literal["python", "json", "yaml"]): The format in which to save the data.
                Only 'python' is currently supported.

        Raises:
            NotImplementedError: If 'json' or 'yaml' modes are selected.
        """
        with h5py.File(self.filename, "a") as f:
            # Handle project overwriting
            if overwrite_project:
                self._overwrite_project(f, model.metadata.project)

            # Create or retrieve project group
            project_group = self._get_or_create_group(f, model.metadata.project)
            self._update_attributes(project_group, model.metadata.__dict__)

            # Create or retrieve model group
            model_group = self._get_or_create_group(project_group, model.version)

            # Handle dataset overwriting
            if overwrite_datasets:
                self._overwrite_datasets(model_group)

            # Save data based on the mode
            if mode == "python":
                self._save_pickled(model_group, "input", model)
                if result:
                    self._save_pickled(model_group, "output", result)
            else:
                raise NotImplementedError(f"Mode '{mode}' is not yet implemented.")

    def load(
        self,
        project: str,
        model: str | None = None,
        load_results: bool = False,
        mode: Literal["python", "json", "yaml"] = "python",
    ) -> dict[str, tuple["Model", Union["Result", None]]]:
        """
        Load a single model or all models within a specific project.

        Parameters:
            project (str): The project name.
            model (str, optional): The model name. If None, loads all models in the project.
            load_results (bool): Whether to load results along with the model.
            mode (Literal["python", "json", "yaml"]): The format in which to load the data.
                Only 'python' is currently supported.

        Returns:
            dict[str, tuple[Model, Union[Result, None]]]: A dictionary of loaded models and their results.

        Raises:
            NotImplementedError: If 'json' or 'yaml' modes are selected.
        """
        loaded_models = {}

        with h5py.File(self.filename, "r") as f:
            # Validate that the project exists
            if project not in f:
                msg = f"Project '{project}' does not exist in the HDF5 file."
                raise KeyError(msg)

            # Get all models in the project
            all_models = self.list_models[project]

            if mode == "python":
                # Load all models if no specific model is provided
                if model is None:
                    for item in all_models:
                        loaded_models[item] = self._load_pickled(
                            group=f[project], name=item, load_results=load_results
                        )
                else:
                    # Validate that the specific model exists
                    if model not in all_models:
                        msg = f"Model '{model}' does not exist in project '{project}'."
                        raise KeyError(msg)
                    loaded_models[model] = self._load_pickled(
                        group=f[project], name=model, load_results=load_results
                    )
            else:
                msg = f"Mode '{mode}' is not yet implemented."
                raise NotImplementedError(msg)

        # Update internal models dictionary and return loaded models
        self.models.update(loaded_models)
        return loaded_models

    def delete(self, project: str, model: str | None = None):
        """Delete a single model or all models within a specific project.

        !!! warning

            Use this method only for small deletions, as for now it does not
            perform repacking of the HDF5 file (the objects are deleted but the
            disk space is not freed). For large deletions, consider creating a
            new HDF5 file and saving only the models you want to keep.

        Parameters:
            project (str): The project name.
            model (str): The model name.
        """
        with h5py.File(self.filename, "a") as f:
            if model is None:
                try:
                    del f[project]
                except KeyError:
                    logger.warning(f"Project {project} does not exist.")
            else:
                try:
                    del f[project][model]
                except KeyError:
                    logger.warning(
                        f"Model {model} does not exist in project {project}."
                    )

    def _overwrite_datasets(self, group):
        """Delete all datasets within a group."""
        for key in list(group.keys()):
            try:
                del group[key]
                logger.info(f"Deleted dataset {key} in group {group.name}")
            except KeyError:
                logger.warning(f"Failed to delete dataset {key} in group {group.name}")

    def _overwrite_project(self, file, project_name):
        """Delete a project group from the HDF5 file."""
        try:
            del file[project_name]
            logger.info(f"Deleted project {project_name}")
        except KeyError:
            logger.warning(f"Project {project_name} does not exist.")

    def _save_plain_string(self, group, name, data):
        """Save data as a plain string dataset."""

    def _save_pickled(self, group, name, data):
        """Save data as a pickled dataset."""
        try:
            pickle_data = pickle.dumps(data)
            group.create_dataset(name, data=np.void(pickle_data))
            logger.info(f"Saved {name} to {group.name}")
        except ValueError as e:
            logger.warning(
                f"Failed to create dataset {name} in {group.name}. Error: {e}"
            )

    def _update_attributes(self, group, attributes):
        """Update attributes of an HDF5 group."""
        sanitized_attrs = {k: v for k, v in attributes.items() if v is not None}
        group.attrs.update(sanitized_attrs)

    def _load_pickled(
        self, group: h5py.Group, name: str, load_results: bool
    ) -> tuple["Model", Union["Result", None]]:
        """
        Load a pickled model and optionally its results from an HDF5 group.

        Parameters:
            group (h5py.Group): The HDF5 group containing the model data.
            name (str): The name of the model to load.
            load_results (bool): Whether to load results along with the model.

        Returns:
            tuple[Model, Union[Result, None]]: The loaded model and its results (if applicable).
        """
        try:
            # Load pickled input data
            pickle_in = group[name]["input"][()].tobytes()
            model: Model = pickle.loads(pickle_in)

            # Optionally load pickled output data
            result: Result = None

            if load_results and "output" in group[name]:
                pickle_out = group[name]["output"][()].tobytes()
                result = pickle.loads(pickle_out)

            logger.info(f"Loaded model '{name}' from group '{group.name}'.")
            return model, result
        except Exception:
            logger.exception(f"Failed to load model '{name}' from group '{group.name}'")
            return None, None



# pyright: reportInvalidTypeForm=false

from typing import Literal

import pandera.pandas as pa
from pandera.typing import Series

from pyswap.core.basemodel import BaseTableModel
from pyswap.core.valueranges import DVSRANGE, UNITRANGE, YEARRANGE

__all__ = [
    "AMAXTB",
    "CFTB",
    "CO2AMAXTB",
    "CO2EFFTB",
    "CO2TRATB",
    "CROPROTATION",
    "DMGRZTB",
    "DMMOWDELAY",
    "DMMOWTB",
    "DTSMTB",
    "FLTB",
    "FOTB",
    "FRTB",
    "FSTB",
    "GCTB",
    "KYTB",
    "LSDATB",
    "LSDBTB",
    "MRFTB",
    "RDCTB",
    "RDRRTB",
    "RDRSTB",
    "RDTB",
    "RFSETB",
    "RLWTB",
    "SHORTINTERVALMETEODATA",
    "SLATB",
    "TMNFTB",
    "TMPFTB",
    "WRTB",
]  #  TODO: needs update?

# %% ++++++++++++++++++++++++++++ CROP TABLES ++++++++++++++++++++++++++++

crop_tables = [
    "DATEHARVEST",
    "RDTB",
    "RDCTB",
    "GCTB",
    "CFTB",
    "INTERTB",
    "KYTB",
    "MRFTB",
    "WRTB",
    "CROPROTATION",
    "DTSMTB",
    "SLATB",
    "AMAXTB",
    "TMPFTB",
    "TMNFTB",
    "RFSETB",
    "FRTB",
    "FLTB",
    "FSTB",
    "FOTB",
    "RDRRTB",
    "RDRSTB",
    "DMGRZTB",
    "LSDATB",
    "LSDBTB",
    "RLWTB",
    "DMMOWTB",
    "DMMOWDELAY",
    "IRRIGEVENTS",
    "TC1TB",
    "TC2TB",
    "TC3TB",
    "TC4TB",
    "TC7TB",
    "TC8TB",
    "DC1TB",
    "DC2TB",
    "CO2EFFTB",
    "CO2TRATB",
    "CO2AMAXTB",
    "LSDA",
]


class DATEHARVEST(BaseTableModel):
    """Date of harvest

    Attributes:
        DATEHARVEST (Series[pa.DateTime]): Date of harvest.
    """

    DATEHARVEST: Series[pa.DateTime]


class RDTB(BaseTableModel):
    """Rooting Depth [0..1000 cm, R], as a function of development stage [0..2 -, R].

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        RD (Series[float]): Rooting depth of the crop.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    RD: Series[float] = pa.Field(ge=0.0, le=100.0)


class RDCTB(BaseTableModel):
    """List root density [0..100 cm/cm3, R] as function of relative rooting depth [0..1 -, R]

    Attributes:
        RRD (Series[float]): Relative rooting depth of the crop.
        RDENS (Series[float]): Root density of the crop.

    """

    RRD: Series[float] = pa.Field(ge=0.0, le=100.0)
    RDENS: Series[float] = pa.Field(**UNITRANGE)


class GCTB(BaseTableModel):
    """Leaf Area Index [0..12 (m2 leaf)/(m2 soil), R], as function of dev. stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        LAI (Series[float]): Leaf Area Index of the crop.
    """

    DVS: Series[float] = pa.Field(**DVSRANGE)
    LAI: Series[float] = pa.Field(ge=0.0, le=12.0)


class CFTB(BaseTableModel):
    """Crop factor [0..2 [-], R], as function of dev. stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        DNR (Series[float]): Day number.
        CF (Series[float]): Crop factor.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    CF: Series[float] | None


class CHTB(BaseTableModel):
    """Crop Height [0..1.d4 cm, R], as function of dev. stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        DNR (Series[float]): Day number.
        CH (Series[float]): Crop height.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    # # CF was added for compatibility with example grass files in original SWAP distribution
    # # that are used for testing this package. CF is only stated but not used there. I
    # removed it because it was conflicting with other settings. Input files for the examples
    # should be adjusted.
    CH: Series[float] | None


class INTERTB(BaseTableModel):
    """Interception parameters for closed forest canopies (SWINTER=2).

    Attributes:
        T (Series[int]): Time [0..366 d, R].
        PFREE (Series[float]): Free throughfall coefficient [0..1 -, R].
        PSTEM (Series[float]): Stem flow coefficient [0..1 -, R].
        SCANOPY (Series[float]): Storage capacity of canopy [0..10 cm, R].
        AVPREC (Series[float]): Average rainfall intensity [0..100 cm/d, R].
        AVEVAP (Series[float]): Average evaporation intensity during rainfall from a wet canopy [0..10 cm/d, R].
    """

    T: Series[float] = pa.Field(ge=0, le=366)
    PFREE: Series[float] = pa.Field(ge=0.0, le=1.0)
    PSTEM: Series[float] = pa.Field(ge=0.0, le=1.0)
    SCANOPY: Series[float] = pa.Field(ge=0.0, le=10.0)
    AVPREC: Series[float] = pa.Field(ge=0.0, le=100.0)
    AVEVAP: Series[float] = pa.Field(ge=0.0, le=10.0)


class KYTB(BaseTableModel):
    """Yield response factor [0..5 -, R], as function of dev. stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        KY (Series[float]): Yield response factor of the crop.
    """

    DVS: Series[float] = pa.Field(**DVSRANGE)
    KY: Series[float] = pa.Field(ge=0.0, le=5.0)


class MRFTB(BaseTableModel):
    """Ratio root total respiration / maintenance respiration [1..5.0 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        MAX_RESP_FACTOR (Series[float]): Ratio root total respiration / maintenance respiration.
    """

    DVS: Series[float] = pa.Field(**DVSRANGE)
    MAX_RESP_FACTOR: Series[float] = pa.Field(ge=1.0, le=5.0)


class WRTB(BaseTableModel):
    """dry weight of roots at soil surface [0..10 kg/m3, R], as a function of development stage [0..2 -,R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        W_ROOT_SS (Series[float]): Dry weight of roots at soil surface.
    """

    DVS: Series[float] = pa.Field(**DVSRANGE)
    W_ROOT_SS: Series[float] = pa.Field(ge=0.0, le=10.0)


class CROPROTATION(BaseTableModel):
    """Crop rotation settings

    Attributes:
        CROPSTART (Series[pa.DateTime]): Start date of the crop.
        CROPEND (Series[pa.DateTime]): End date of the crop.
        CROPFIL (Series[str]): Crop file name.
        CROPTYPE (Series[int]): Crop module type

            * 1 - simple
            * 2 - detailed, WOFOST general
            * 3 - detailed, WOFOST grass
    """

    CROPSTART: Series[pa.DateTime]
    CROPEND: Series[pa.DateTime]
    CROPFIL: Series[str]
    CROPTYPE: Series[int] = pa.Field(ge=1, le=3)


# WOFOST-specific tables
class DTSMTB(BaseTableModel):
    """increase in temperature sum [0..60 oC, R] as function of daily average temperature [0..100 oC, R]

    Attributes:
        TAV (Series[float]): Daily average temperature.
        DTSM (Series[float]): Increase in temperature sum.
    """

    TAV: Series[float] = pa.Field(ge=0.0, le=100.0)
    DTSM: Series[float] = pa.Field(ge=0.0, le=60.0)


class SLATB(BaseTableModel):
    """Specific leaf area [0..1 ha/kg, R] as function of crop development stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        SLA (Series[float]): Leaf area.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    SLA: Series[float] = pa.Field(ge=0.0, le=1.0)


class AMAXTB(BaseTableModel):
    """maximum CO2 assimilation rate [0..100 kg/ha/hr, R] as function of development stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        AMAX (Series[float]): Maximum CO2 assimilation rate.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    AMAX: Series[float] = pa.Field(ge=0.0, le=100.0)


class TMPFTB(BaseTableModel):
    """reduction factor of AMAX [-, R] as function of average day temperature [-10..50 oC, R]

    Attributes:
        TAVD (Series[float]): Minimum temperature.
        TMPF (Series[float]): Reduction factor of AMAX.
    """

    TAVD: Series[float] = pa.Field(ge=-10.0, le=50.0)
    TMPF: Series[float] = pa.Field(ge=0.0, le=1.0)


class TMNFTB(BaseTableModel):
    """reduction factor of AMAX [-, R] as function of minimum day temperature [-10..50 oC, R]

    Attributes:
        TMNR (Series[float]): Minimum temperature.
        TMNF (Series[float]): Reduction factor of AMAX.
    """

    TMNR: Series[float] = pa.Field(ge=-10.0, le=50.0)
    TMNF: Series[float] = pa.Field(ge=0.0, le=1.0)


class RFSETB(BaseTableModel):
    """reduction factor of senescence [-, R] as function of development stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        RFSE (Series[float]): Reduction factor of senescence.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    RFSE: Series[float] = pa.Field(ge=0.0, le=1.0)


class FRTB(BaseTableModel):
    """fraction of total dry matter increase partitioned to the roots [kg/kg, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        FR (Series[float]): Fraction of total dry matter increase partitioned to the roots.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    FR: Series[float] = pa.Field(ge=0.0, le=1.0)


class FLTB(BaseTableModel):
    """fraction of total above ground dry matter increase partitioned to the leaves [kg/kg, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        FL (Series[float]): Fraction of total above ground dry matter increase partitioned to the leaves.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    FL: Series[float] = pa.Field(ge=0.0, le=1.0)


class FSTB(BaseTableModel):
    """fraction of total above ground dry matter increase partitioned to the stems [kg/kg, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        FS (Series[float]): Fraction of total above ground dry matter increase partitioned to the stems.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    FS: Series[float] = pa.Field(ge=0.0, le=1.0)


class FOTB(BaseTableModel):
    """fraction of total above ground dry matter increase partitioned to the storage organs [kg/kg, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        FO (Series[float]): Fraction of total above ground dry matter increase partitioned to the storage organs.
    """

    DVS: Series[float] = pa.Field(**DVSRANGE)
    FO: Series[float] = pa.Field(ge=0.0, le=1.0)


class RDRRTB(BaseTableModel):
    """relative death rates of roots [kg/kg/d] as function of development stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        RDRR (Series[float]): Relative death rates of roots.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    RDRR: Series[float] = pa.Field(ge=0.0)


class RDRSTB(BaseTableModel):
    """relative death rates of stems [kg/kg/d] as function of development stage [0..2 -, R]

    Attributes:
        DVS (Series[float]): Development stage of the crop.
        RDRS (Series[float]): Relative death rates of stems.
    """

    DVS: Series[float] | None = pa.Field(**DVSRANGE)
    DNR: Series[float] | None = pa.Field(**YEARRANGE)
    RDRS: Series[float] = pa.Field(ge=0.0)


class DMGRZTB(BaseTableModel):
    """threshold of above ground dry matter [0..1d6 kg DM/ha, R] to trigger grazing as function of daynumber [1..366 d, R]

    Attributes:
        DNR (Series[float]): Day number.
        DMGRZ (Series[float]): Dry matter growth rate of roots.
    """

    DNR: Series[float] = pa.Field(**YEARRANGE)
    DMGRZ: Series[float] = pa.Field(ge=0.0, le=1.0e6)


class LSDATB(BaseTableModel):
    """Actual livestock density of each grazing period

    !!! note

        total number of periods should be equal to number of periods in SEQGRAZMOW

    Attributes:
        SEQNR (Series[int]): number of the sequence period with mowing/grazing [0..366 d, I]
        LSDA (Series[float]): Actual Live Stock Density of the grazing period [0.0..1000.0 LS/ha, R]
    """

    SEQNR: Series[int] = pa.Field(**YEARRANGE)
    LSDA: Series[float] = pa.Field(ge=0.0, le=1000.0)


class LSDBTB(BaseTableModel):
    """Relation between livestock density, number of grazing days and dry matter uptake

    Attributes:
        LSDB (Series[float]): Basic Live Stock Density [0.0..1000.0 LS/ha, R]
        DAYSGRAZING (Series[float]): Maximum days of grazing [0.0..366.0 d, R]
        UPTGRAZING (Series[float]): Dry matter uptake by grazing [0.0..1000.0 kg/ha, R] (kg/ha DM)
        LOSSGRAZING (Series[float]): Dry matter loss during grazing due to droppings and treading [0.0..1000.0 kg/ha, R] (kg/ha DM)
    """

    LSDb: Series[float] = pa.Field(ge=0.0, le=1000.0)
    DAYSGRAZING: Series[float] = pa.Field(**YEARRANGE)
    UPTGRAZING: Series[float] = pa.Field(ge=0.0, le=1000.0)
    LOSSGRAZING: Series[float] = pa.Field(ge=0.0, le=1000.0)


class RLWTB(BaseTableModel):
    """rooting depth RL [0..5000 cm, R] as function of root weight RW [0..5000 kg DM/ha, R]

    Attributes:
        RW (Series[float]): rooting depth
        RL (Series[float]): root weight
    """

    RW: Series[float] = pa.Field(ge=0.0, le=5000.0)
    RL: Series[float] = pa.Field(ge=0.0, le=5000.0)


class DMMOWTB(BaseTableModel):
    """List threshold of above ground dry matter [0..1d6 kg DM/ha, R] to trigger mowing as function of daynumber [1..366 d, R]

    !!! note

        maximum 20 records


    Attributes:
        DNR (Series[float]): Day number.
        DMMOW (Series[float]): threshold of above ground dry matter [0..1d6 kg DM/ha, R]
    """

    DNR: Series[float] = pa.Field(**YEARRANGE)
    DMMOW: Series[float] = pa.Field(ge=0.0, le=1.0e6)


class DMMOWDELAY(BaseTableModel):
    """Relation between dry matter harvest [0..1d6 kg/ha, R] and days of delay in regrowth [0..366 d, I] after mowing

    Attributes:
        DMMOWDELAY (Series[float]): Dry matter harvest [0..1d6 kg/ha, R]
        DAYDELAY (Series[int]): days of delay in regrowth [0..366 d, I]
    """

    DMMOWDELAY: Series[float] = pa.Field(ge=0.0, le=1.0e6)
    DAYDELAY: Series[int] = pa.Field(**YEARRANGE)


class IRRIGEVENTS(BaseTableModel):
    """information for each fixed irrigation event.

    Attributes:
        IRDATE (Series[datetime]):date of irrigation.
        IRDEPTH (Series[float]): amount of water [0..1000 mm, R].
        IRCONC (Series[float]): concentration of irrigation water [0..1000 mg/cm3, R].
        IRTYPE (Series[int]): type of irrigation

            * 0 - sprinkling
            * 1 - surface

    """

    IRDATE: Series[pa.DateTime]
    IRDEPTH: Series[float] | None = pa.Field(default=None, ge=0.0, le=1000.0)
    IRCONC: Series[float] = pa.Field(ge=0.0, le=1000.0)
    IRTYPE: Series[int] = pa.Field(ge=0, le=1)


class TC1TB(BaseTableModel):
    """tc1tb option table"""

    DVS_TC1: Series[float] = pa.Field(ge=0.0, le=2.0)
    TREL: Series[float] = pa.Field(ge=0.0, le=1.0)


class TC2TB(BaseTableModel):
    """tc2tb option table"""

    DVS_TC2: Series[float] = pa.Field(ge=0.0, le=2.0)
    RAW: Series[float] = pa.Field(ge=0.0, le=1.0)


class TC3TB(BaseTableModel):
    """tc3tb option table"""

    DVS_TC3: Series[float] = pa.Field(ge=0.0, le=2.0)
    TAW: Series[float] = pa.Field(ge=0.0, le=1.0)


class TC4TB(BaseTableModel):
    """tc4tb option table"""

    DVS_TC4: Series[float] = pa.Field(ge=0.0, le=2.0)
    DWA: Series[float] = pa.Field(ge=0.0, le=500.0)


class TC7TB(BaseTableModel):
    """tc7tb option table"""

    DVS_TC7: Series[float] = pa.Field(ge=0.0, le=2.0)
    HCRI: Series[float] = pa.Field(ge=-1000.0, le=-100.0)


class TC8TB(BaseTableModel):
    """tc8tb option table"""

    DVS_TC8: Series[float] = pa.Field(ge=0.0, le=2.0)
    TCRI: Series[float] = pa.Field(ge=0.0, le=1.0)


class DC1TB(BaseTableModel):
    DVS_DC1: Series[float]
    DI: Series[float]


class DC2TB(BaseTableModel):
    DVS_DC2: Series[float]
    FID: Series[float]


class CO2EFFTB(BaseTableModel):
    """Correction factor light use efficiency for change in CO2 concentration."""

    CO2PPM: Series[float]
    FACTOR: Series[float]


class CO2TRATB(BaseTableModel):
    """Correction factor maximum transpiration rate for change in CO2 concentration."""

    CO2PPM: Series[float]
    FACTOR: Series[float]


class CO2AMAXTB(BaseTableModel):
    """Correction factor assimilation rate for change in CO2 concentration."""

    CO2PPM: Series[float]
    FACTOR: Series[float]


class VERNRTB(BaseTableModel):
    """Vernalization rate as a function of average air temperature."""

    TAV: Series[float]
    VERNRATE: Series[float]


class LSDA(BaseTableModel):
    """Actual livestock density of each grazing period

    !!! note

        total number of periods should be equal to number of periods in SEQGRAZMOW

    Attributes:
        SEQNR (Series[int]): number of the sequence period with mowing/grazing [0..366 d, I]
        LSDA (Series[float]): Actual Live Stock Density of the grazing period [0.0..1000.0 LS/ha, R]
    """

    SEQNR: Series[int] = pa.Field(**YEARRANGE)
    LSDA: Series[float] = pa.Field(ge=0.0, le=1000.0)


# %% ++++++++++++++++++++++++++++ METEO TABLES ++++++++++++++++++++++++++++

meteo_tables = [
    "DAILYMETEODATA",
    "SHORTINTERVALMETEODATA",
    "DETAILEDRAINFALL",
    "RAINFLUX",
]


class DAILYMETEODATA(BaseTableModel):
    """Format detailed daily meteo data.

    TODO:
    format decimals in the variables.
    """

    STATION: Series[str]
    DD: Series[str]
    MM: Series[str]
    YYYY: Series[str]
    RAD: Series[float]
    TMIN: Series[float]
    TMAX: Series[float]
    HUM: Series[float]
    WIND: Series[float]
    RAIN: Series[float]
    ETREF: Series[float]
    WET: Series[float]


class SHORTINTERVALMETEODATA(BaseTableModel):
    Date: Series[pa.DateTime]
    Record: Series[int] = pa.Field(ge=1, le=10)
    Rad: Series[float]
    Temp: Series[float]
    Hum: Series[float]
    Wind: Series[float]
    Rain: Series[float]


class DETAILEDRAINFALL(BaseTableModel):
    Station: Series[str]
    Day: Series[int]
    Month: Series[int]
    Year: Series[int]
    Time: Series[float]
    Amount: Series[float]


class RAINFLUX(BaseTableModel):
    TIME: Series[float] = pa.Field(**YEARRANGE)
    RAINFLUX: Series[float] = pa.Field(ge=0, le=1000.0)


# %% ++++++++++++++++++++++++++++ SOILWATER TABLES ++++++++++++++++++++++++++++

soilwater_tables = [
    "INIPRESSUREHEAD",
    "MXPONDTB",
    "SOILPROFILE",
    "SOILHYDRFUNC",
    "SOILTEXTURES",
    "INITSOILTEMP",
]


class INIPRESSUREHEAD(BaseTableModel):
    """Initial pressure head [cm, R] as a function of soil layer [1..N, I].

    Attributes:
        ZI: Series[int]: soil depth [-1.d5..0 cm, R].
        H: Series[float]: Initial soil water pressure head [-1.d10..1.d4 cm, R].
    """

    ZI: Series[float] = pa.Field(ge=-1.0e5, le=0.0)
    H: Series[float] = pa.Field(ge=-1.0e10, le=1.0e4)


class MXPONDTB(BaseTableModel):
    """minimum thickness for runoff PONDMXTB [0..1000 cm, R] as function of time

    Attributes:
        DATEPMX: Series[pa.DateTime]: Date of the ponding threshold for runoff.
        PONDMXTB: Series[float]: Minimum thickness for runoff.
    """

    DATEPMX: Series[pa.DateTime]
    PONDMXTB: Series[float]


class SOILPROFILE(BaseTableModel):
    """Vertical discretization of soil profile

    Attributes:
        ISUBLAY: Series[int]: number of sub layer, start with 1 at soil surface [1..MACP, I].
        ISOILLAY: Series[int]: number of soil physical layer, start with 1 at soil surface [1..MAHO, I].
        HSUBLAY: Series[float]: height of sub layer [0..1.d4 cm, R].
        HCOMP: Series[float]: height of compartments in the sub layer [0.0..1000.0 cm, R].
        NCOMP: Series[int]: number of compartments in the sub layer (Mind NCOMP = HSUBLAY/HCOMP) [1..MACP, I].
    """

    ISOILLAY: Series[int] = pa.Field(ge=1)
    ISUBLAY: Series[int] = pa.Field(ge=1)
    HSUBLAY: Series[float] = pa.Field(ge=0.0, le=1.0e4)
    HCOMP: Series[float] = pa.Field(ge=0.0, le=1.0e3)
    NCOMP: Series[int] = pa.Field(ge=1)


class SOILHYDRFUNC(BaseTableModel):
    """Soil hydraulic functions table.

        !!! warning
            ALFAW required only when the hysteresis option is set to 1 or 2. This column is set as optional column and (for now) is not checked.

    Attributes:
        ORES (Series[float]): Residual water content [0..1 cm3/cm3, R]
        OSAT (Series[float]): Saturated water content [0..1 cm3/cm3, R]
        ALFA (Series[float]): Parameter alfa of main drying curve [0.0001..100 /cm, R]
        NPAR (Series[float]): Parameter n [1.001..9 -, R]
        LEXP (Series[float]): Exponent in hydraulic conductivity function [-25..25 -, R]
        KSATFIT (Series[float]): Fitting parameter Ksat of hydraulic conductivity function [1.d-5..1d5 cm/d, R]
        H_ENPR (Series[float]): Air entry pressure head [-40.0..0.0 cm, R]
        KSATEXM (Series[float]): Measured hydraulic conductivity at saturated conditions [1.d-5..1d5 cm/d, R]
        BDENS (Series[float]): Dry soil bulk density [100..1d4 mg/cm3, R]
        ALFAW (Optional[Series[float]]): Alfa parameter of main wetting curve in case of hysteresis [0.0001..100 /cm, R]
    """

    ORES: Series[float] = pa.Field(ge=0.0, le=1.0)
    OSAT: Series[float] = pa.Field(ge=0.0, le=1.0)
    ALFA: Series[float] = pa.Field(ge=0.0001, le=100.0)
    NPAR: Series[float] = pa.Field(ge=1.001, le=9.0)
    LEXP: Series[float] = pa.Field(ge=-25.0, le=25.0)
    KSATFIT: Series[float] = pa.Field(ge=1.0e-5, le=1.0e5)
    H_ENPR: Series[float] = pa.Field(ge=-40.0, le=0.0)
    KSATEXM: Series[float] = pa.Field(ge=1.0e-5, le=1.0e5)
    BDENS: Series[float] = pa.Field(ge=100.0, le=1.0e4)
    ALFAW: Series[float] | None = pa.Field(ge=0.0001, le=100.0)


# %% ++++++++++++++++++++++++++++ HEAT FLOW TABLES ++++++++++++++++++++++++++++


class SOILTEXTURES(BaseTableModel):
    """Table for soil textures.

    Attributes:
        PSAND (float): Depth of soil layer [cm, R]
        PSILT (float): Sand content [g/g mineral parts, R]
        PCLAY (float): Clay content [g/g mineral parts, R]
        ORGMAT (float): Organic matter content [g/g dry soil, R]
    """

    PSAND: float
    PSILT: float
    PCLAY: float
    ORGMAT: float


class INITSOILTEMP(BaseTableModel):
    """Table for initial soil temperature.

    Attributes:
        ZH (float): Depth of soil layer [cm, R]
        TSOIL (float): Initial temperature [oC, R]
    """

    ZH: float = pa.Field(ge=-100000, le=0)
    TSOIL: float = pa.Field(ge=-50, le=50)


# %% ++++++++++++++++++++++++++++ BOUNDARY TABLES ++++++++++++++++++++++++++++

boundary_tables = [
    "GWLEVEL",
    "QBOT2",
    "HAQUIF",
    "QBOT4",
    "QTAB",
    "HBOT5",
    "DATET",
    "CSEEPARR",
    "INISSOIL",
    "MISC",
]


class GWLEVEL(BaseTableModel):
    """Table for groundwater levels.

    Attributes:
        DATE (Series[pa.DateTime]): Date of the groundwater level.
        GWLEVEL (Series[float]): Groundwater level.
    """

    DATE1: Series[pa.DateTime]
    GWLEVEL: Series[float]


class QBOT2(BaseTableModel):
    """Table for bottom boundary flow.

    Attributes:
        DATE2 (Series[pa.DateTime]): Date of the bottom boundary flow.
        QBOT2 (Series[float]): Bottom boundary flow.
    """

    DATE2: Series[pa.DateTime]
    QBOT2: Series[float]


class HAQUIF(BaseTableModel):
    """Table for aquifer thickness.

    Attributes:
        DATE3 (Series[pa.DateTime]): Date of the aquifer thickness.
        HAQUIF (Series[float]): Aquifer thickness.
    """

    DATE3: Series[pa.DateTime]
    HAQUIF: Series[float]


class QBOT4(BaseTableModel):
    """Table for bottom boundary flow.

    Attributes:
        DATE4 (Series[pa.DateTime]): Date of the bottom boundary flow.
        QBOT4 (Series[float]): Bottom boundary flow.
    """

    DATE4: Series[pa.DateTime]
    QBOT4: Series[float]


class QTAB(BaseTableModel):
    """Table for height of the water table.

    Attributes:
        HTAB (Series[pa.DateTime]): Date of the height of the water table.
        HTABLE (Series[float]): Height of the water table.
    """

    HTAB: Series[float]
    QTAB: Series[float]


class HBOT5(BaseTableModel):
    """Table for bottom compartment pressure head.

    Attributes:
        DATE6 (Series[pa.DateTime]): Date of the bottom compartment pressure head.
        HBOT5 (Series[float]): Bottom compartment pressure head.
    """

    DATE5: Series[pa.DateTime]
    HBOT5: Series[float]


class DATET(BaseTableModel):
    """Table for time.

    Attributes:
        DATE7 (Series[pa.DateTime]): Date of the time.
        TIME (Series[float]): Time.
    """

    DATET: Series[pa.DateTime]
    TBOT: Series[float]


class CSEEPARR(BaseTableModel):
    """Table for seepage.

    Attributes:
        DATE8 (Series[pa.DateTime]): Date of the seepage.
        CSEEPARR (Series[float]): Seepage.
    """

    DATEC: Series[pa.DateTime]
    CSEEPARR: Series[float]


class INISSOIL(BaseTableModel):
    """Table for capillary rise.

    Attributes:
        DATE9 (Series[pa.DateTime]): Date of the capillary rise.
        CML (Series[float]): Capillary rise.
    """

    ZC: Series[float]
    CML: Series[float]


class MISC(BaseTableModel):
    """Table for miscellaneous.

    Attributes:
        DATE10 (Series[pa.DateTime]): Date of the miscellaneous.
        MISC (Series[float]): Miscellaneous.
    """

    LDIS: Series[float]
    KF: Series[float]
    DECPOT: Series[float]
    FDEPTH: Series[float]


# %% ++++++++++++++++++++++++++++ DRAINAGE TABLES ++++++++++++++++++++++++++++

drainage_tables = [
    "DRNTB",
    "DRAINAGELEVELTOPPARAMS",
    "DATOWLTB1",
    "DATOWLTB2",
    "DATOWLTB3",
    "DATOWLTB4",
    "DATOWLTB5",
    "SECWATLVL",
    "MANSECWATLVL",
    "QWEIR",
    "QWEIRTB",
    "PRIWATLVL",
    "QDRNTB",
]


class DRNTB(BaseTableModel):
    """Drainage characteristics table.

    Attributes:
        LEV (Series[int]): Drainage level [1..5, I].
        SWDTYP (Series[int]): Type of drainage medium: 1 = drain tube, 2 = open channel.
        L (Series[float]): Drain spacing [1..100000 m, R].
        ZBOTDRE (Series[float]): Level of drainage medium bottom [-10000..0 cm, R].
        GWLINF (Series[float]): Groundwater level influence [-10000..200 cm, R].
        RDRAIN (Series[float]): Drainage resistance [10..1d5 d, R].
        RINFI (Series[float]): Infiltration resistance [0..1d5 d, R].
        RENTRY (Series[float]): Entry resistance [0..1000 d, R].
        REXIT (Series[float]): Exit resistance [0..1000 d, R].
        WIDTHR (Series[float]): Width of the drainage medium [0..1000 cm, R].
        TALUDR (Series[float]): Talud of the drainage medium [0..1000 cm, R].
    """

    LEV: Series[int] = pa.Field(ge=1, le=5)
    SWDTYP: Series[Literal[0, 1]]
    L: Series[float] = pa.Field(ge=1.0, le=100000.0)
    ZBOTDRE: Series[float]
    GWLINF: Series[float] = pa.Field(ge=-1000.0, le=0.0)
    RDRAIN: Series[float] = pa.Field(ge=1.0, le=100000.0)
    RINFI: Series[float] = pa.Field(ge=1.0, le=100000.0)
    RENTRY: Series[float] = pa.Field(ge=0.0, le=100.0)
    REXIT: Series[float] = pa.Field(ge=0.0, le=100.0)
    WIDTHR: Series[float] = pa.Field(ge=0.0, le=10000.0)
    TALUDR: Series[float] = pa.Field(ge=0.01, le=5.0)


class DRAINAGELEVELTOPPARAMS(BaseTableModel):
    """Drainage level top parameters table.

    I couldn't find the description of this table. It occurs in the swa.dra file
    of the Hupselbrook example, but not in the templates.

    Attributes:
        SWTOPDISLAY (Series[Literal[0, 1]]): Switch for each drainage level to distribute drainage flux vertically with a given position of the top of the model discharge layers [Y=1, N=0]
        ZTOPDISLAY (Series[float]): Array with depth of top of model discharge layer for each drain level [-10000.0..0.0, cm, R]
        FTOPDISLAY (Series[float]): Array with factor of top of model discharge layer for each drain level [0.0..1.0, -, R]
    """

    SWTOPDISLAY: Series[Literal[0, 1]]
    ZTOPDISLAY: Series[float] = pa.Field(ge=-10000.0, le=0.0)
    FTOPDISLAY: Series[float] = pa.Field(ge=0.0, le=1.0)


class DATOWLTB1(BaseTableModel):
    """Table for drainage water level.

    Attributes:
        DATOWL1 (Series[pa.DateTime]): Date of the drainage water level.
        WLEVEL (Series[float]): Drainage water level.
    """

    DATOWL1: Series[pa.DateTime]
    LEVEL1: Series[float]


class DATOWLTB2(BaseTableModel):
    """Table for drainage water level.

    Attributes:
        DATOWL2 (Series[pa.DateTime]): Date of the drainage water level.
        LEVEL2 (Series[float]): Drainage water level.
    """

    DATOWL2: Series[pa.DateTime]
    LEVEL2: Series[float]


class DATOWLTB3(BaseTableModel):
    """Table for drainage water level.

    Attributes:
        DATOWL3 (Series[pa.DateTime]): Date of the drainage water level.
        LEVEL3 (Series[float]): Drainage water level.
    """

    DATOWL3: Series[pa.DateTime]
    LEVEL3: Series[float]


class DATOWLTB4(BaseTableModel):
    """Table for drainage water level.

    Attributes:
        DATOWL4 (Series[pa.DateTime]): Date of the drainage water level.
        LEVEL4 (Series[float]): Drainage water level.
    """

    DATOWL4: Series[pa.DateTime]
    LEVEL4: Series[float]


class DATOWLTB5(BaseTableModel):
    """Table for drainage water level.

    Attributes:
        DATOWL5 (Series[pa.DateTime]): Date of the drainage water level.
        LEVEL5 (Series[float]): Drainage water level.
    """

    DATOWL5: Series[pa.DateTime]
    LEVEL5: Series[float]


class SECWATLVL(BaseTableModel):
    DATE2: Series[pa.DateTime]
    WLS: Series[float]


class MANSECWATLVL(BaseTableModel):
    IMPER_4B: Series[float]
    IMPEND: Series[pa.DateTime]
    SWMAN: Series[float]
    WSCAP: Series[float]
    WLDIP: Series[float]
    INTWL: Series[float]


class QWEIR(BaseTableModel):
    IMPER_4C: Series[float]
    HBWEIR: Series[float]
    ALPHAW: Series[float]
    BETAW: Series[float]


class QWEIRTB(BaseTableModel):
    IMPER_4D: Series[float]
    IMPTAB: Series[float]
    HTAB: Series[float]
    QTAB: Series[float]


class PRIWATLVL(BaseTableModel):
    DATE1: Series[pa.DateTime]
    WLP: Series[float]


class QDRNTB(BaseTableModel):
    QDRAIN: Series[float]
    GWL: Series[float]


# %% ++++++++++++++++++++++++++++ GENERAL SETTINGS TABLES ++++++++++++++++++++++++++++

general_settings_tables = ["OUTDATIN", "OUTDAT"]


class OUTDATIN(BaseTableModel):
    """OUTDATIN table

    Attributes:
        OUTDAT: Series[str]: Name of the output file.
    """

    OUTDATIN: Series[pa.DateTime]


class OUTDAT(BaseTableModel):
    """OUTDAT table

    Attributes:
        OUTDAT: Series[str]: Name of the output file.
    """

    OUTDAT: Series[pa.DateTime]



"""Command Line Interface for pySWAP.

This is a prototype subpackage for potential enhancement of pyswap's
functionality. CLI tools can be very helpful in automating some tasks, like
loading databases or classic SWAP models.

!!! note

    At the moment only creating project structure was prototyped. More
    functionality will be added in the future if users express such need.


Example:

    ```cmd
    pyswap init --notebook  # creates the project structure with a template .ipynb file.
    pyswap init --script  # creates the project structure with a .py file.
    ```

After running the script, you will see the following folder created:

```cmd
test project
├── README
├── __init__.py
├── data
├── models
│   ├── __init__.py
│   └── main.ipynb
└── scripts
    └── __init__.py
```

The `__init__.py files are added to create a module structure. Now when you create a python file in scripts with some helper functions,
you can import those functions to the main model script or notebook and use it there.`

```python
from ..scripts.helper_module import helper_function

var = helper_function(**kwargs)
```

By default, a git repository is also created along with the project structure.
"""



# mypy: disable-error-code="attr-defined, no-any-return, index"
# attr-defined error showed in when the .update method was called in the wofost mixin.
# the easiest way to fix it would be to just check for the PySWAPBaseModel
# instance, but that would create circular import issue and I did not want to
# play with that. It's not a priority to fix for now.
# no-any-return error was raised in the get_origin method. This was not a
# priority to fix. Same as the index error; it complained that the FieldInfo
# object is not indexable, but in fact in this case it should be, because of the
# Optional or Annotated types.

"""Reusable mixins enhancing functionality of specific PySWAPBaseModel.

To keep the main PySWAPBaseModel class and the components library clean and
focused, mixins are used to add additional functionality to the classes that
need it. The concept of the mixins was inspired by the Django framework and it
really helps to keep the code clean and organized.

Should more functionality be needed in the future for one or more classes, it
should be implemented as a mixin and then inherited by the classes that need it.

Classes:

    FileMixin: Custom saving functionality for models that need file I/O.
    SerializableMixin: Converting a model to a SWAP-formatted string.
    YAMLValidatorMixin: Validating parameters using external YAML rules.
    WOFOSTUpdateMixin: Interface for the WOFOST crop parameters database for
        pySWAP.
"""

from pathlib import Path
from typing import Any, Literal, Self, Union, get_args, get_origin

from pydantic import BaseModel, PrivateAttr, model_serializer, model_validator
from pydantic.fields import FieldInfo

from pyswap.core.defaults import VALIDATIONRULES
from pyswap.core.io.io_yaml import load_yaml
from pyswap.core.io.processors import TableProcessor
from pyswap.log import logging

logger = logging.getLogger(__name__)


class FileMixin:
    """Custom saving functionality for models that need file I/O.

    !!! note:

        The _extension attribute should be set in the class that inherits
        this mixin. It is recommended that pydantic's PrivateAttr is used to
        hide this attribute from the user.

    Methods:
        save_file: Saves a string to a file.
    """

    def save_file(
        self,
        string: str,
        fname: str,
        path: Path,
    ) -> None:
        """Saves a string to a file.

        The extension should now be provided in each class inheriting this
        mixin as a private attribute.

        Parameters:
            string: The string to be saved to a file.
            fname: The name of the file.
            path: The path where the file should be saved.
        """

        if not hasattr(self, "_extension"):
            msg = "The _extension attribute should be set."
            raise AttributeError(msg)

        ext = self._extension
        fname = f"{fname}.{ext}" if ext else fname

        with open(f"{path}/{fname}", "w", encoding="ascii") as f:
            f.write(string)

        logger.info(f"{fname} saved successfully.")

        return None


class SerializableMixin(BaseModel):
    """Converting a model to a SWAP-formatted string.

    This mixin is only inherited by classes that directly serialize to a
    SWAP-formatted string. The assumptions are that the inheriting classes:

    - do not contain nested classes.
    - if the class contains nested classes it should either use Subsection field
        types or override the `model_string()` method.

    Methods:
        if_is_union_type: Check if the field type is a Union type.
        is_annotated_exception_type: Check if the attribute type is Table,
            Arrays, or ObjectList.
        serialize_model: Override the default serialization method.
        model_string: Concatenate the formatted strings from dictionary to
            one string.
    """

    def if_is_union_type(self, field_info: FieldInfo) -> dict | None:
        """Check if the field type is a Union type.

        If it is, look for the json_schema_extra attribute in the field_info
        of the first argument of the Union type. If it is not found, return
        None. It was necessary in cases of, for example, optional classes like
        Union[Table, None].

        Parameters:
            field_info (FieldInfo): The FieldInfo object of the field.
        """

        field_type = field_info.annotation

        if get_origin(field_type) is Union:
            union_args = get_args(field_type)
            args = get_args(union_args[0])

            field_info = [item for item in args if isinstance(item, FieldInfo)]

            if not field_info:
                return None

            # Only return the json_schema_extra attribute. This is used in some
            # cases to pass addotional information from the serializer in
            # pyswap.core.fields module to the model_dump.
            return field_info[0].json_schema_extra
        return None

    def is_annotated_exception_type(self, field_name: str) -> bool:
        """Check if the attribute type is Table, Arrays, or ObjectList.

        For Table, Arrays, and ObjectList types True is returned, ensuring a
        separate serialization path.

        First try to assign the json_schema_extra from a Union type. If that
        fails, assign the json_schema_extra from the field_info. If the
        json_schema_extra is None, return False.
        """
        # Every special field will have a FieldInfo object
        _class = type(self)
        field_info = _class.model_fields.get(field_name, None)

        if field_info is None:
            return False

        json_schema_extra = (
            self.if_is_union_type(field_info) or field_info.json_schema_extra
        )

        if json_schema_extra is None:
            return False

        return json_schema_extra.get("is_annotated_exception_type", False)

    @model_serializer(when_used="json", mode="wrap")
    def serialize_model(self, handler: Any):
        """Override the default serialization method.

        In the intermediate step, a dictionary is created with SWAP formatted
        strings.
        """
        result = {}
        validated_self = handler(self)
        for field_name, field_value in validated_self.items():
            if self.is_annotated_exception_type(field_name):
                result[field_name] = field_value
            else:
                result[field_name] = f"{field_name.upper()} = {field_value}"
        return result

    def model_string(
        self, mode: Literal["str", "list"] = "string", **kwargs
    ) -> str | list[str]:
        """Concatenate the formatted strings from dictionary to one string.


        !!! note:
            By alias is True, because in some cases, particularily in the case
            of CropSettings, the WOFOST names of parameters in the database were
            different from those used in SWAP. This allows those parameters to
            be properly matched, yet serialized properly in SWAP input files.

        Parameters:
            mode (Literal["str", "list]): The output format.
            kwargs (dict): Additional keyword arguments passed to `model_dump()`.
        """
        dump = self.model_dump(
            mode="json", exclude_none=True, by_alias=True, **kwargs
        ).values()

        if mode == "list":
            return list(dump)
        else:
            return "\n".join(dump)


class YAMLValidatorMixin(BaseModel):
    """A mixin class that provides YAML-based validation for models.

    Initially, pySWAP had model serializers on each model component class which
    had a number of assertions to validate the parameters (i.e., require
    parameters rlwtb and wrtmax if swrd = 3). This created chaos
    in the code, and since none of it was used by inspection tools anyways, it
    was decided to leave the validation logic in the code and move the rules to
    a separate YAML file.

    Methods:
        validate_parameters: Validate parameters against required rules.
        validate_with_yaml: Pydantic validator executing validation logic.
    """

    _validation: bool = PrivateAttr(default=False)

    @staticmethod
    def validate_parameters(
        switch_name: str, switch_value: str, params: dict, rules: dict
    ):
        """Validate parameters against required rules.

        This method reads the rules for the model from the YAML file and checks
        if the required parameters are present. If not, it raises a ValueError.

        ```yaml
        SaltStress: # <--- Model name
            swsalinity:  # <--- Switch name (switch_name)
                1:  # <--- Switch value (switch_value)
                - saltmax  # <---| Required parameters
                - saltslope  # <--|
                2:
                - salthead
        ```

        Parameters:
            switch_name (str): The name of the switch (e.g., 'swcf').
            switch_value (Any): The value of the switch (e.g., 1 or 2).
            params (dict): Dictionary of parameters to check.
            rules (dict): Dictionary with validation rules.

        Raises:
            ValueError: If required parameters are missing.
        """

        required_params = rules.get(switch_name, {}).get(switch_value, [])

        if not required_params:
            return  # No rules for this switch value

        missing_params = [
            param for param in required_params if params.get(param) is None
        ]

        if missing_params:
            msg = f"The following parameters are required for {switch_name}={switch_value}: {', '.join(missing_params)}"
            raise ValueError(msg)

    @model_validator(mode="after")
    def validate_with_yaml(self) -> Self:
        """Pydantic validator executing validation logic.

        All validators defined on a model run on model instantiation. This
        method makes sure that YAML validation is postponed until the
        _validation parameter (required on all classes inheriting this mixin) is
        set to True. This state is done when all the required parameters are
        presumed to be set, e.g., when the user tries to run the model.
        """

        if not self._validation:
            return self

        rules = VALIDATIONRULES.get(self.__class__.__name__, {})

        logger.debug(f"Validating {self.__class__.__name__} with rules: {rules}")

        for switch_name in rules:
            switch_value = getattr(self, switch_name, None)
            if switch_value is not None:  # Only validate if the switch is set
                self.validate_parameters(
                    switch_name, switch_value, self.__dict__, rules
                )

        self._validation = False
        return self


class YAMLUpdateMixin:
    """Interface for the WOFOST crop parameters database for pySWAP.

    This mixin should be inherited by classes that share parameters with the
    WOFOST crop database.
    """

    def _process_parameters(
        self, params: dict, source_name: str, grass: bool = False
    ) -> dict:
        """Process parameters by converting tables/arrays using TableProcessor.

        Args:
            params: Dictionary of parameters to process
            source_name: Name of the parameter source (for logging)

        Returns:
            Processed parameters dictionary with tables converted to dataframes
        """
        params_copy = params.copy()
        tp = TableProcessor()

        for name, value in params.items():
            processed = None

            # Handle YAML dict format (two-column tables)
            if isinstance(value, dict) and len(value) == 2:
                keys = list(value.keys())
                logger.debug(
                    f"Processing YAML dict format for parameter: {name} with keys: {keys}"
                )
                if len(keys) == 2:
                    array_data = [
                        [x, y]
                        for x, y in zip(value[keys[0]], value[keys[1]], strict=True)
                    ]
                    processed = tp.process(
                        "array", data=array_data, columns=name, grass=grass
                    )

            # Handle list of lists format (array parameters)
            elif isinstance(value, list) and value and isinstance(value[0], list):
                processed = tp.process("array", data=value, columns=name, grass=grass)

            # Update or remove parameter based on processing result
            if processed is not None:
                params_copy.update(processed)
                logger.debug(f"Processed parameter from {source_name}: {name}")
            elif isinstance(value, list | dict) and not isinstance(value, str):
                # Remove unmatched complex parameters (but keep scalars)
                params_copy.pop(name, None)
                logger.warning(
                    f"Failed to process parameter from {source_name}: {name}, removing from update"
                )

        return params_copy

    def update_from_wofost(self) -> None:
        """Update the model with the WOFOST variety settings."""

        # parameters attribute returns a dictionary with the key-value pairs and
        # tables as list of lists. Before updating, the tables should be
        # created.
        if not hasattr(self, "wofost_variety"):
            msg = "The model does not have the WOFOST variety settings."
            raise AttributeError(msg)

        variety_params = self.wofost_variety.parameters
        logger.debug(f"Updating from WOFOST variety parameters: {variety_params}")

        processed_params = self._process_parameters(variety_params, "WOFOST")
        self.update(processed_params, inplace=True)

    def update_from_yaml(self, yaml_path: str | Path, grass: bool = False) -> None:
        """Update the model with parameters from a YAML file.

        Parameters:
            yaml_path (str | Path): Path to the YAML file containing parameters.
        """

        yaml_path = Path(yaml_path)
        if not yaml_path.exists():
            msg = f"YAML file not found: {yaml_path}"
            raise FileNotFoundError(msg)

        yaml_content = load_yaml(yaml_path)
        logger.debug(f"Loaded YAML content from: {yaml_path}")

        # Extract parameters from the YAML structure
        # Assuming the YAML has a structure like: CropParameters -> SWAPInput -> parameters
        params = None
        if (
            isinstance(yaml_content, dict)
            and "CropParameters" in yaml_content
            and isinstance(yaml_content["CropParameters"], dict)
            and "SWAPInput" in yaml_content["CropParameters"]
        ):
            params = yaml_content["CropParameters"]["SWAPInput"]
        elif isinstance(yaml_content, dict):
            # If it's a flat dictionary, use it directly
            params = yaml_content
        else:
            structure_info = str(type(yaml_content).__name__)
            if hasattr(yaml_content, "keys"):
                structure_info = (
                    f"{structure_info} with keys: {list(yaml_content.keys())}"
                )
            msg = f"Could not find parameters in YAML structure: {structure_info}"
            raise ValueError(msg)

        logger.debug(f"Extracted parameters from YAML: {list(params.keys())}")

        processed_params = self._process_parameters(params, "YAML", grass)
        self.update(processed_params, inplace=True)



"""Value ranges for pydantic Field objects used in pyswap validation.

Pydantic's Field() object is used to define metadata and constraints for model
fields. This module contains commonly used ranges of values for fields in pySWAP
classes.

Value ranges:
    UNITRANGE (dict): Range of values between 0.0 and 1.0.
    YEARRANGE (dict): Range of values for year (0 <= x <= 366).
    DVSRANGE (dict): Range of values for development stage (0 <= x <= 2).
"""

UNITRANGE = {"ge": 0.0, "le": 1.0}
"""Range of values between 0.0 and 1.0."""

YEARRANGE = {"ge": 0, "le": 366}
"""Range of values for year (0 <= x <= 366)."""

DVSRANGE = {"ge": 0.0, "le": 2.0}
"""Range of values for development stage (0 <= x <= 2)."""



import pandas as pd
import pytest
from pandera.errors import SchemaError

import pyswap as psp
import pyswap.components.meteorology as meteo
import pyswap.testcase as testcase


def test_metfile_from_csv():
    # Test correct csv file
    table_test = meteo.metfile_from_csv(
        metfil="test_meteo_1.csv", csv_path="tests/test_meteo/test_meteo_1.csv"
    )
    # expected output
    table_exp = pd.read_csv("tests/test_meteo/test_meteo_1.csv")
    table_exp = table_exp.rename(
        {col: col.upper() for col in table_exp.columns}, axis=1
    )
    # Compare
    pd.testing.assert_frame_equal(
        table_test.content,
        table_exp,
        check_dtype=False,
    )

    # Test csv file with missing column
    with pytest.raises(SchemaError) as exc_info:
        table_test = meteo.metfile_from_csv(
            metfil="test_meteo_2.csv", csv_path="tests/test_meteo/test_meteo_2.csv"
        )
    assert "not in dataframe. Columns in dataframe:" in str(exc_info.value)

    # Test csv file with string value instead of float
    with pytest.raises(SchemaError) as exc_info:
        table_test = meteo.metfile_from_csv(
            metfil="test_meteo_3.csv", csv_path="tests/test_meteo/test_meteo_3.csv"
        )
    assert "Error while coercing" in str(exc_info.value)

    # Test csv file with float value instead of string
    with pytest.raises(SchemaError) as exc_info:
        table_test = meteo.metfile_from_csv(
            metfil="test_meteo_4.csv", csv_path="tests/test_meteo/test_meteo_4.csv"
        )
    assert "Error while coercing" in str(exc_info.value)

    # Test if station field starts with '
    table_test = meteo.metfile_from_csv(
        metfil="test_meteo_5.csv", csv_path="tests/test_meteo/test_meteo_5.csv"
    )
    assert table_test.content["STATION"].iloc[0].startswith("'")


def test_metfile_from_knmi():
    # Test default variables
    table_test = meteo.metfile_from_knmi(
        metfil="test_knmi",
        stations="260",
        start="20200103",
        end="20200103",
    ).content

    # expected output
    table_exp = pd.read_csv("tests/test_meteo/test_knmi.csv")
    table_exp["STATION"] = table_exp["STATION"].astype(str)
    table_exp.loc[:, "STATION"] = table_exp.STATION.apply(
        lambda x: f"'{x}'" if not str(x).startswith("'") else x
    )
    table_exp["DD"] = table_exp["DD"].astype(str)
    table_exp["MM"] = table_exp["MM"].astype(str)
    table_exp["YYYY"] = table_exp["YYYY"].astype(str)
    # Compare
    pd.testing.assert_frame_equal(
        table_test,
        table_exp,
        check_dtype=False,
        rtol=1e-2,
    )


def test_hupsel_with_knmi_input():
    model = testcase.get("hupselbrook")

    metfile = psp.components.meteorology.metfile_from_knmi(
        metfil="knmitest.met", stations="260", start="2002-01-01", end="2004-12-31"
    )

    meteo = psp.components.meteorology.Meteorology(
        metfile=metfile,
        lat=52.0,
        alt=2,  # m
        swetr=0,  # Penman-Monteith
        angstroma=0.25,  # TODO
        angstromb=0.50,  # TODO
        swmetdetail=0,  # daily data
        swdivide=1,  # divide E and T using PM
        swrain=0,  # Use only daily rainfall amounts
        swetsine=0,  # Distribute Tp and Ep over the day using a sine wave
        altw=2.0,  # wind, m
    )

    model.meteorology = meteo

    model.run("./", silence_warnings=True)



from pyswap.testcase.load_dataset import (
    get_path,
    load_met,
    load_observations,
    view_resources,
)
from pyswap.testcase.run_testcase import get

__all__ = [
    "get",
    "get_path",
    "load_met",
    "load_observations",
    "view_resources",
]



import pyswap as psp


def _make_simple_test_model():
    # Generate empty model instance
    ml = psp.Model()

    # Metadata
    meta = psp.components.Metadata(
        author="markvdbrink",
        institution="Wageningen University",
        email="mark.vandenbrink@wur.nl",
        project="02_DrainageDevelopment",
        swap_ver="4.2",
    )
    ml.metadata = meta

    # General model settings
    simset = psp.components.simsettings.GeneralSettings(
        swscre=1,  # Print simulation progression
        swerror=1,  # Print error messages
        tstart="2000-01-01",
        tend="2000-12-31",
        nprintday=1,  # number of output times a day
        swmonth=1,  # monthly output
        # period=1,  # daily output, swmonth=0
        swyrvar=0,  # specify output for .bal and .blc file, unnecessary buth required by SWAP
        datefix="31 12",  # specify output for .bal and .blc file, unnecessary buth required by SWAP
        extensions=["csv", "csv_tz"],  # type of output files
        inlist_csv=[
            "watbal",
            "etterms",
            "crop",
        ],  # check SWAPtools for more options
        inlist_csv_tz=["wc", "h", "o2top", "rwu"],
    )
    ml.generalsettings = simset

    richardsettings = psp.components.simsettings.RichardsSettings(
        swkmean=1,  # Use weighted mean of hydraulic conductivity
        swkimpl=0,  # Do not update hydraulic conductivity within iteration
        dtmin=1e-6,  # minimum time step [d]
        dtmax=0.04,  # maximum time step [d]
        gwlconv=100.0,  # Maximum difference of groundwater level between iterations [cm]
        critdevh1cp=1e-3,  # Maximum relative difference in pressure heads per compartment
        critdevh2cp=1e-2,  # Maximum absolute difference in pressure heads per compartment
        critdevponddt=1e-4,  # Maximum water balance error of ponding layer [cm]
        maxit=30,  # Maximum number of iterations
        maxbacktr=5,  # Maximum number of backtracking cycles within an iteration cycle
    )
    ml.richardsettings = richardsettings

    # Meteorology
    ## metfile
    metfile = psp.components.meteorology.metfile_from_knmi(
        metfil="260.met",
        stations=["260"],
        start="2000-01-01",
        end="2000-12-31",
    )

    meteo = psp.components.meteorology.Meteorology(
        metfile=metfile,
        lat=52.0,
        alt=2,  # [m]
        altw=2.0,  # altitude of wind measurements [m]
        swetr=0,  # Use Penman-Monteith, not ETref
        angstroma=0.25,  # Angstrom coefficient a (Allen et al, 1998)
        angstromb=0.50,  # Angstrom coefficient b (Allen et al, 1998)
        swmetdetail=0,  # daily data
        swdivide=1,  # divide E and T using PM
        swrain=0,  # Use only daily rainfall amounts, not intensity TODO
        swetsine=0,  # Do not distribute Tp and Ep over the day using a sine wave
    )

    ml.meteorology = meteo

    # Initial soil moisture content
    soilmoisture = psp.components.soilwater.SoilMoisture(
        swinco=2,  # Initial soil moisture in static equilibirum with initial groundwater level
        gwli=-85.0,  # Initial groundwater level [cm]
    )
    ml.soilmoisture = soilmoisture

    # Surface flow
    surfaceflow = psp.components.soilwater.SurfaceFlow(
        swrunon=0,  # No runon
        swpondmx=0,  # Constant ponding threshold
        pondmx=0.2,  # Minimum ponding thickness for runoff to occur [cm]
        rsro=0.5,  # resistance to runoff [d]
        rsroexp=1.0,  # exponent of drainage equation for runoff
    )
    ml.surfaceflow = surfaceflow

    # Soil evaporation
    evaporation = psp.components.soilwater.Evaporation(
        swcfbs=0,  # Do not use soil factor to calculate Epot from ETref, unneccesary but required by SWAP
        swredu=2,  # Darcy + Boesten/Stroosnijder for reduction potential soil evaporation top layer
        cofredbo=0.54,  # Reduction factor Boesten/Stroosnijder for potential soil evaporation
        rsoil=150.0,  # Minimum soil resistance to evaporation [s/m]
    )
    ml.evaporation = evaporation

    soil_profile = psp.components.soilwater.SOILPROFILE.create(
        {
            "ISUBLAY": [1, 2, 3, 4],
            "ISOILLAY": [1, 1, 2, 2],
            "HSUBLAY": [10.0, 20.0, 30.0, 140.0],
            "HCOMP": [1.0, 5.0, 5.0, 10.0],
            "NCOMP": [10, 4, 6, 14],
        }
    )

    soil_hydraulic_functions = psp.components.soilwater.SOILHYDRFUNC.create(
        {
            "ORES": [0.01, 0.02],
            "OSAT": [0.42, 0.38],
            "ALFA": [0.0276, 0.0213],
            "NPAR": [1.491, 1.951],
            "KSATFIT": [12.52, 12.68],
            "LEXP": [-1.060, 0.168],
            "ALFAW": [0.0542, 0.0426],
            "H_ENPR": [0.0, 0.0],
            "KSATEXM": [12.52, 12.68],
            "BDENS": [1315.0, 1315.0],
        }
    )

    soilprofile = psp.components.soilwater.SoilProfile(
        swsophy=0,
        soilprofile=soil_profile,
        swhyst=0,
        tau=0.2,
        soilhydrfunc=soil_hydraulic_functions,
        swmacro=0,
    )

    ml.soilprofile = soilprofile

    # Bottom boundary condition
    bottom_boundary = psp.components.boundary.BottomBoundary(
        swbbcfile=0,  # Do not specify in separate file
        swbotb=6,  # Bottom flux equals zero
    )
    ml.bottomboundary = bottom_boundary

    # Crop settings
    ## Crop preparation settings
    maize_prep = psp.components.crop.Preparation(
        swprep=0, swsow=0, swgerm=0, dvsend=3.0, swharv=0
    )

    ## Crop development settings
    dvs = [0.0, 0.3, 0.5, 0.7, 1.0, 1.4, 2.0]

    ## Crop height
    maize_chtb = psp.components.crop.CHTB.create(
        {
            "DVS": dvs,
            "CH": [1.0, 15.0, 40.0, 140.0, 170.0, 180.0, 175.0],
        }
    )

    ## Root density as function of depth
    maize_rdctb = psp.components.crop.RDCTB.create(
        {
            "RRD": [0.0, 1.0],
            "RDENS": [1.0, 0.0],
        }
    )

    ### Get WOFOST parameters for maize from database
    db_wofost = psp.db.WOFOSTCropDB()
    maize_wofostparams = db_wofost.load_crop_file("maize").get_variety(
        "Grain_maize_201"
    )

    ### Define other parameters
    maize_cropdev_settings = psp.components.crop.CropDevelopmentSettingsWOFOST(
        # Wofost parameters for maize
        wofost_variety=maize_wofostparams,
        # Evaporation parameters
        swcf=2,  # Use crop height
        chtb=maize_chtb,
        albedo=0.2,  # Albedo of the crop
        rsc=167.0,  # Minimum canopy resistance [s/m]
        rsw=0.0,  # Canopy resistance of intercepted water [s/m]
        # Initial values
        laiem=0.04836,  # LAI at emergence [m2/m2]
        # Green surface area
        ssa=0.0,  # Specific leaf area [ha/kg]
        # Assimilation
        kdif=0.6,  # Extinction coefficient for diffuse light
        kdir=0.75,  # Extinction coefficient for direct light
        eff=0.45,  # Light use efficiency [ka/ha/hr/(J m2 s)]
        # Root development
        rdc=100.0,  # Maximum root depth [cm]
        swrd=2,  # Root depth depends on maximum daily increase
        swdmi2rd=0,  # Rooting depth increase depends on assimilate availability
        rdctb=maize_rdctb,
    )

    ### Get parameters from WOFOST database
    maize_cropdev_settings.update_from_wofost()

    ## Test update function -> validation is triggered
    maize_cropdev_settings = maize_cropdev_settings.update({"RSC": 0.0})

    ## Oxygen stress
    maize_ox_stress = psp.components.crop.OxygenStress(
        swoxygen=1,  # Feddes stress function
        swwrtnonox=0,  # Do not stop root development when anaerobic conditions occur
        hlim1=-15.0,  # no RWU at higher h
        hlim2u=-30.0,  # optimal RWU at lower h, upper layer
        hlim2l=-30.0,  # optimal RWU at lower h, lower layer
    )

    ## Drought stress
    maize_dr_stress = psp.components.crop.DroughtStress(
        swdrought=1,  # Feddes stress function
        hlim3h=-325.0,  # h below which RWU reduction starts at high Tpot
        hlim3l=-600.0,  # h below which RWU reduction starts at low Tpot
        hlim4=-8000.0,  # No RWU at h below this value
        adcrh=0.5,  # High Tpot [cm]
        adcrl=0.1,  # Low Tpot [cm]
    )

    ## Interception
    maize_interception = psp.components.crop.Interception(
        swinter=1,  # method for agricultural crops
        cofab=0.25,  # max interception amount [cm]
    )

    ## Maize crop file
    crpmaize = psp.components.crop.CropFile(
        name="maizes",
        prep=maize_prep,
        cropdev_settings=maize_cropdev_settings,
        oxygenstress=maize_ox_stress,
        droughtstress=maize_dr_stress,
        interception=maize_interception,
    )

    ## Crop rotation
    croprotation = psp.components.crop.CROPROTATION.create(
        {
            "CROPSTART": ["2000-05-01"],
            "CROPEND": ["2000-10-15"],
            "CROPFIL": ["'maizes'"],
            "CROPTYPE": [2],
        }
    )

    # ## No irrigation
    # irrigation = psp.components.crop.Irrigation(
    #     swirr=0,  # No irrigation
    # )

    ## Final crop file
    crop = psp.components.crop.Crop(
        swcrop=1,  # Simulate crop
        rds=40.0,  # Maximum rooting depth [cm]
        croprotation=croprotation,
        cropfiles={"maizes": crpmaize},
    )
    ml.crop = crop

    # Lateral drainage
    ## Freeboard over time
    datowltb1 = psp.components.drainage.DATOWLTB1.create(
        {
            "DATOWL1": ["2000-01-01", "2000-12-31"],
            "LEVEL1": [-80.0, -80.0],
        }
    )

    ## Drainage flux
    flux = psp.components.drainage.Flux(
        drares1=100,  # d
        infres1=1e5,  # d
        swallo1=3,  # only drainage is allowed
        l1=50.0,  # drain spacing [m], to allow for vertical distribution
        zbotdr1=-85.0,  # bottom of drainage medium [cm]
        swdtyp1=2,  # type is open channel
        datowltb1=datowltb1,  # drainage level
    )

    ## Drainage file
    dra = psp.components.drainage.DraFile(
        dramet=3,  # Use resistance
        swdivd=1,  # Calculate vertical distribution of drainage flux
        cofani=[1.0, 1.0],  # anisotropy factor
        swdislay=0,  # No adjustment top layer
        nrlevs=1,  # Number of levels
        swtopnrsrf=0,  # No adjustment bottom discharge layer
        swintfl=0,  # No interflow in highest drainage level
        fluxes=flux,
    )

    # Drainage defition
    drainage = psp.components.drainage.Drainage(
        swdra=1,  # Simulate drainage with basic routine
        drafile=dra,
    )
    ml.lateraldrainage = drainage

    # Heat flow
    # Soil textures for each physical layer
    soiltextures = psp.components.transport.SOILTEXTURES.create(
        {
            "PSAND": [0.87, 0.89],
            "PSILT": [0.1, 0.08],
            "PCLAY": [0.03, 0.03],
            "ORGMAT": [0.057, 0.022],
        }
    )

    # Initial soil temperatures; source: WWL
    initsoiltemp = psp.components.transport.INITSOILTEMP.create(
        {
            "ZH": [-10.0, -40.0, -70.0, -95.0],
            "TSOIL": [12.0, 12.0, 10.0, 9.0],
        }
    )

    heatflow = psp.components.transport.HeatFlow(
        swhea=1,  # Simulate heat flow in soil
        swcalt=2,  # Use the numerical method
        swtopbhea=1,  # Air temperature is top boundary condition
        swbotbhea=1,  # No heat flux in bottom boundary
        initsoiltemp=initsoiltemp,
        soiltextures=soiltextures,
    )
    ml.heatflow = heatflow

    return ml


if __name__ == "__main__":
    ml = _make_simple_test_model()
    ml.run()



# mypy: disable-error-code="override, func-returns-value, call-overload, operator, no-any-return"
# Override error is related to the validate method, must have somethiong to do
# with Pydantic.
# func-returns-value is raised on the model builder, because it does not return
# anything. This is not a priority to fix.
# the operator was returned on one of the checks (swdra comparison), which is
# ok. It is validated prior to reaching that point.
# no-any-return is raised on the run_parallel function, because it returns a
# list of results. This is not a priority to fix.
# securing the subprocess call does not seem like a priority in this project
"""Building, running and parsing the results of a SWAP model run.

When the Model class begun to grow, it was clear that it needed to be refactored
into a more modular structure. The functionality to build environment, run and
parse result has been abstracted to 3 classes, focusing the main (and exposed to
the user) Model class focused on the model components and their interactions.
The four classes in this module are:

Classes:

    ModelBuilder: Class responsible for building the model components.
    ModelRunner: Class responsible for running the model.
    ResultReader: Class responsible for parsing the model results.
    Model: Main class that runs the SWAP model.
"""

from __future__ import annotations

import logging
import os
import subprocess
import tempfile
from multiprocessing import Pool
from pathlib import Path
from typing import Literal

from pandas import DataFrame, read_csv, to_datetime
from pydantic import Field, PrivateAttr, model_validator

from pyswap.components.boundary import BottomBoundary
from pyswap.components.crop import Crop
from pyswap.components.drainage import Drainage
from pyswap.components.irrigation import FixedIrrigation
from pyswap.components.metadata import Metadata
from pyswap.components.meteorology import Meteorology
from pyswap.components.simsettings import GeneralSettings, RichardsSettings
from pyswap.components.soilwater import (
    Evaporation,
    SnowAndFrost,
    SoilMoisture,
    SoilProfile,
    SurfaceFlow,
)
from pyswap.components.transport import HeatFlow, SoluteTransport
from pyswap.core.basemodel import PySWAPBaseModel
from pyswap.core.fields import Subsection
from pyswap.core.io.io_ascii import open_ascii
from pyswap.db.co2concentration import CO2concentration
from pyswap.model.result import Result
from pyswap.utils.executables import get_swap
from pyswap.utils.mixins import FileMixin, SerializableMixin

logger = logging.getLogger(__name__)

__all__ = ["Model", "run_parallel"]


class ModelBuilder:
    """Building model components.

    Attributes:
        model (Model): The model to build.
        tempdir (str): The temporary directory to store the input files.

        write_inputs: Write the input files to the temporary directory.
    """

    def __init__(self, model: Model, tempdir: str):
        self.model = model
        self.tempdir = tempdir

    def get_inputs(self) -> dict:
        """Get the inpup files in a dictionary."""
        inputs = {}

        inputs["swp"] = self.model.swp
        if self.model.lateraldrainage.swdra in [1, 2]:
            inputs["dra"] = self.model.lateraldrainage.drafile.dra
        if self.model.crop.cropfiles:
            inputs["crop"] = self.model.crop.cropfiles
        if self.model.meteorology.metfile:
            inputs["met"] = self.model.meteorology.met
        if self.model.fixedirrigation.swirgfil == 1:
            inputs["irg"] = self.model.fixedirrigation.irg
        if self.model.bottomboundary.swbbcfile == 1:
            inputs["bbc"] = self.model.bottomboundary.bbc

        return inputs

    def write_inputs(self) -> None:
        """Write the input files to the temporary directory."""
        logger.info("Preparing files...")

        self.model.write_swp(self.tempdir)

        if self.model.lateraldrainage.swdra in [1, 2]:
            self.model.lateraldrainage.write_dra(self.tempdir)
        if self.model.crop.cropfiles:
            self.model.crop.write_crop(self.tempdir)
            # Check if co2correction is used in any of the cropfiles
            crops_co2corrections = [
                cropfile.co2correction is not None
                for cropfile in self.model.crop.cropfiles.values()
            ]
            # If so, copy the co2correction file to the tempdir
            if sum(crops_co2corrections) > 0:
                co2db = CO2concentration()
                period = [
                    self.model.generalsettings.tstart,
                    self.model.generalsettings.tend,
                ]
                co2db.write_co2(self.tempdir, period)
        if self.model.meteorology.metfile:
            self.model.meteorology.write_met(self.tempdir)
        if self.model.fixedirrigation.swirgfil == 1:
            self.model.fixedirrigation.write_irg(self.tempdir)
        if self.model.bottomboundary.swbbcfile == 1:
            self.model.bottomboundary.write_bbc(self.tempdir)

        return self


class ModelRunner:
    """Class responsible for running the model.

    In the run method, the ResultReader is utilized to abstract the parsing of
    the model results.

    Attributes:
        model (Model): The model to run.

    Methods:
        run_swap: Run the SWAP executable.
        raise_swap_warning: Raise a warning.
        run: Main function that runs the model
    """

    def __init__(self, model: Model):
        self.model = model

    @staticmethod
    def run_swap(tempdir: Path) -> tuple[str, str, int]:
        """Run the SWAP executable.

        Run the exacutable in the tempdirectory and pass the newline to the
        stdin when the executable asks for input (upon termination).

        get_swap will now automatically install the SWAP executable if it is not found and
        return the path to the executable.

        Parameters:
            tempdir (Path): The temporary directory where the executable
                is stored.

        Returns:
            tuple[str, str, int]: stdout, stderr, and return code

        Raises:
            RuntimeError: If the executable cannot be run (e.g., wrong architecture,
                not executable, corrupt file).
        """
        swap_path = get_swap(verbose=False, auto_install=True)

        try:
            p = subprocess.Popen(
                swap_path,
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=tempdir,
            )
            stdout, stderr = p.communicate(input=b"\n")
            return stdout.decode(), stderr.decode(), p.returncode
        except OSError as e:
            # This catches issues like: permission denied, exec format error (wrong architecture),
            # file not found, etc.
            msg = (
                f"Failed to execute SWAP binary at {swap_path}.\n"
                f"Error: {e}\n"
                f"This may indicate:\n"
                f"  - Wrong executable architecture (e.g., Windows binary on Linux)\n"
                f"  - Corrupted executable file\n"
                f"  - Missing executable permissions\n"
                f"  - Incompatible SWAP version"
            )
            raise RuntimeError(msg) from e
        except Exception as e:
            msg = f"Unexpected error while running SWAP executable: {e}"
            raise RuntimeError(msg) from e

    def raise_swap_warning(self, warnings: list):
        """Log the warnings form the model run.

        Parameters:
            warnings (list): The warnings from the model run parsed with the
                ResultReaded.
        """
        for message in warnings:
            logger.warning(message)

    def run(self, path: str | Path, silence_warnings: bool = False) -> Result:
        """Main function that runs the model.

        First ModelBuilder is used to prepare the environment for the model run.
        Second, the SWAP executable is run and the decoded result passed from
        the executable is parsed using the ResultReader and used to update the
        Result object.

        Parameters:
            path (str | Path): The path to the temporary directory.
            silence_warnings (bool): If True, warnings are not raised.

        Returns:
            Result: The parsed model results.
        """

        with tempfile.TemporaryDirectory(dir=path) as tempdir:
            builder = ModelBuilder(self.model, tempdir)

            builder.write_inputs()

            stdout, stderr, returncode = self.run_swap(tempdir)
            reader = ResultReader(self.model, tempdir)

            if "normal completion" not in stdout.lower():
                # Build a comprehensive error message
                error_parts = ["SWAP model run failed."]

                if returncode != 0:
                    error_parts.append(f"\nReturn code: {returncode}")

                if stderr:
                    error_parts.append(f"\nStderr output:\n{stderr}")

                if stdout:
                    error_parts.append(f"\nStdout output:\n{stdout}")

                # Try to read error file
                err_content = reader.read_error_file()
                if err_content:
                    error_parts.append(f"\nError file (.err):\n{err_content}")

                # Try to read log file
                try:
                    log = reader.read_swap_log()
                    error_parts.append(f"\nLog file (.log):\n{log}")
                except FileNotFoundError:
                    error_parts.append("\nLog file (.log): Not found - executable may have crashed before writing log file")

                # Try to read warning file
                wrn_content = reader.read_warning_file()
                if wrn_content:
                    error_parts.append(f"\nWarning file (.wrn):\n{wrn_content}")

                msg = "\n".join(error_parts)
                logger.error(msg)
                raise RuntimeError(msg)

            logger.info(stdout)

            # --- Handle the results ---
            result: Result = Result()

            log = reader.read_swap_log()
            result.log = log

            warnings = reader.identify_warnings(log)
            result.warning = warnings

            if warnings and not silence_warnings:
                self.raise_swap_warning(warnings=warnings)

            if "csv" in self.model.generalsettings.extensions:
                output = reader.read_csv_output(which="csv")
                result.output.update({"csv": output})

            if "csv_tz" in self.model.generalsettings.extensions:
                output_tz = reader.read_csv_output(which="csv_tz")
                result.output.update({"csv_tz": output_tz})

            ascii_files = reader.read_ascii_output()

            result.output.update(ascii_files)
            return result


class ResultReader:
    """Class responsible for reading the model results.

    Attributes:
        model (Model): The model to read the results from.
        tempdir (str): The temporary directory where the results are stored.

    Methods:
        read_csv_output: Read the csv output.
        read_swap_log: Read the log files.
        identify_warnings: Catch warnings from the log file.
        read_ascii_output: Read all output files that are not in csv format
            as strings.
    """

    def __init__(self, model: Model, tempdir: str):
        self.model: Model = model
        self.tempdir = tempdir

    def read_csv_output(self, which: Literal["csv", "csv_tz"]) -> DataFrame:
        """Read the csv output.

        There are two types of csv output files: csv and csv_tz. They are both
        handle in the same method with mode change.

        Parameters:
            which (str): The type of output file to read.

        Returns:
            DataFrame: The output file as a DataFrame.
        """

        outfil = self.model.generalsettings.outfil
        output_suffix = "_output.csv" if which == "csv" else "_output_tz.csv"
        index_col = "DATETIME" if which == "csv" else "DATE"

        path = Path(self.tempdir, outfil + output_suffix)

        if not path.exists():
            logger.warning(f"Expected output file {path} not found.")
            return DataFrame()

        df = read_csv(path, comment="*", index_col=index_col)
        df.index = to_datetime(df.index)

        return df

    def read_swap_log(self) -> str:
        """Read the log files.

        Returns:
            str: The content of the log file.

        Raises:
            FileNotFoundError: If no log file is found. There should always be
                a log file. If not, something went wrong.
            FileExistsError: If multiple log files are found. Not sure if this
                is possible or not. If so, it should be handled.
        """

        log_files = [
            f for f in Path(self.tempdir).glob("*.log") if f.name != "reruns.log"
        ]

        if len(log_files) == 0:
            msg = "No .log file found in the directory."
            raise FileNotFoundError(msg)

        elif len(log_files) > 1:
            msg = "Multiple .log files found in the directory."
            raise FileExistsError(msg)

        log_file = log_files[0]

        with open(log_file) as file:
            log_content = file.read()

        return log_content

    def read_error_file(self) -> str | None:
        """Read the SWAP error file (.err).

        SWAP writes error messages to a .err file when it encounters
        fatal errors.

        Returns:
            str | None: The content of the error file, or None if not found.
        """
        err_files = list(Path(self.tempdir).glob("*.err"))

        if len(err_files) == 0:
            return None

        err_file = err_files[0]

        try:
            with open(err_file) as file:
                content = file.read()
            # Only return if file has content
            return content.strip() if content.strip() else None
        except Exception:
            return None

    def read_warning_file(self) -> str | None:
        """Read the SWAP warning file (.wrn).

        SWAP writes warning messages to a .wrn file during execution.

        Returns:
            str | None: The content of the warning file, or None if not found.
        """
        wrn_files = list(Path(self.tempdir).glob("*.wrn"))

        if len(wrn_files) == 0:
            return None

        wrn_file = wrn_files[0]

        try:
            with open(wrn_file) as file:
                content = file.read()
            # Only return if file has content
            return content.strip() if content.strip() else None
        except Exception:
            return None

    @staticmethod
    def identify_warnings(log: str) -> list:
        """Catch warnings from the log file.

        This is used by the ModelRunner to raise warnings after the model run.

        Parameters:
            log (str): The log file content.

        Returns:
            list: A list of warnings.
        """
        lines = log.split("\n")
        warnings = [
            line for line in lines if line.strip().lower().startswith("warning")
        ]
        return warnings

    def read_ascii_output(self):
        """Read all output files that are not csv format as strings.

        This method is perhaps a bit oversimplified. In the future, we might
        think about introducing parsers for the different output files. For now,
        we just read them as strings.

        Returns:
            dict (dict): A dictionary of the output strings with extension as key.
        """

        ascii_extensions = [
            ext
            for ext in self.model.generalsettings.extensions
            if ext not in ["csv", "csv_tz"]
        ]

        list_dir = os.listdir(self.tempdir)
        list_dir = [f for f in list_dir if f.endswith(tuple(ascii_extensions))]

        if list_dir:
            dict_files = {
                f.split(".")[1]: open_ascii(Path(self.tempdir, f)) for f in list_dir
            }
            return dict_files
        return {}


class Model(PySWAPBaseModel, FileMixin, SerializableMixin):
    """Main class that runs the SWAP model.

    Even though all sections are set to optional, the model will not run if
    any of the components are missing, except for fixedirrigation, SnowAndFrost,
    RichardsSettings, HeatFlow and SoluteTransport. These components have
    default values that will be used if they are not provided.

    Attributes:
        metadata (Subsection): Metadata of the model.
        version (str): The version of the model (default: "base").
        generalsettings (Subsection): Simulation settings.
        richardsettings (Subsection): Richards settings.
        meteorology (Subsection): Meteorological data.
        crop (Subsection): Crop data.
        fixedirrigation (Subsection): Fixed irrigation settings (default: no irrigation).
        soilmoisture (Subsection): Soil moisture data.
        surfaceflow (Subsection): Surface flow data.
        evaporation (Subsection): Evaporation data.
        soilprofile (Subsection): Soil profile data.
        snowandfrost (Subsection): Snow and frost data (default: no snow and frost).
        lateraldrainage (Subsection): Lateral drainage data.
        bottomboundary (Subsection): Bottom boundary data.
        heatflow (Subsection): Heat flow data.
        solutetransport (Subsection): Solute transport data.

    Methods:
        write_swp: Write the .swp input file.
        validate: Validate the model.
        run: Run the model.
    """

    _validate_on_run: bool = PrivateAttr(default=False)
    _extension = "swp"

    metadata: Subsection[Metadata] | None = Field(default=None, repr=False)
    version: str = Field(exclude=True, default="base")
    generalsettings: Subsection[GeneralSettings] | None = Field(
        default=None, repr=False
    )
    richardsettings: Subsection[RichardsSettings] | None = Field(
        default=RichardsSettings(swkmean=1, swkimpl=0), repr=False
    )
    meteorology: Subsection[Meteorology] | None = Field(default=None, repr=False)
    crop: Subsection[Crop] | None = Field(default=None, repr=False)
    fixedirrigation: Subsection[FixedIrrigation] | None = Field(
        default=FixedIrrigation(swirfix=0), repr=False
    )
    soilmoisture: Subsection[SoilMoisture] | None = Field(default=None, repr=False)
    surfaceflow: Subsection[SurfaceFlow] | None = Field(default=None, repr=False)
    evaporation: Subsection[Evaporation] | None = Field(default=None, repr=False)
    soilprofile: Subsection[SoilProfile] | None = Field(default=None, repr=False)
    snowandfrost: Subsection[SnowAndFrost] | None = Field(
        default=SnowAndFrost(swsnow=0, swfrost=0), repr=False
    )
    lateraldrainage: Subsection[Drainage] | None = Field(default=None, repr=False)
    bottomboundary: Subsection[BottomBoundary] | None = Field(default=None, repr=False)
    heatflow: Subsection[HeatFlow] | None = Field(default=HeatFlow(swhea=0), repr=False)
    solutetransport: Subsection[SoluteTransport] | None = Field(
        default=SoluteTransport(swsolu=0), repr=False
    )

    @property
    def swp(self):
        """The content of the swp file.

        Serialization of Subsection field type has been set in a way that it
        will generate SWAP formatted string when `model_string()` is called on
        the parent class.
        """
        return self.model_string()

    @model_validator(mode="after")
    def validate_missing_components(self):
        """Validate, on run, that all required components are present."""

        if not self._validate_on_run:
            return self

        required_components = [
            "metadata",
            "generalsettings",
            "richardsettings",
            "meteorology",
            "crop",
            "fixedirrigation",
            "soilmoisture",
            "surfaceflow",
            "evaporation",
            "soilprofile",
            "snowandfrost",
            "lateraldrainage",
            "bottomboundary",
            "heatflow",
            "solutetransport",
        ]

        missing_components = [
            comp for comp in required_components if getattr(self, comp) is None
        ]

        if missing_components:
            msg = f"Missing required components: {', '.join(missing_components)}"
            raise ValueError(msg)

        # validate each component
        for comp in required_components:
            getattr(self, comp)

        return self

    @model_validator(mode="after")
    def validate_each_component(self):
        """Validate, on run, that all required components are present."""

        if not self._validate_on_run:
            return self

        for comp in Model.model_fields:
            item = getattr(self, comp)
            if hasattr(item, "validate_with_yaml"):
                item._validation = True
                item.validate_with_yaml()

        return self

    def validate(self):
        """Execute the model validation when `run()` is called.

        This method should probably be refactored. It seems to shadow some
        validation method from Pydantic.
        """

        try:
            self._validate_on_run = True
            self.model_validate(self)
        finally:
            self._validate_on_run = False
            logger.info("Validation successful.")

    def write_swp(self, path: str | Path):
        """Write the .swp input file.

        Parameters:
            path (str | Path): The path to write the file to.
        """
        self.save_file(string=self.swp, path=path, fname="swap")

    def get_inputs(self) -> dict:
        """Get the input files in a dictionary."""
        builder = ModelBuilder(model=self, tempdir=Path.cwd())
        return builder.get_inputs()

    def to_classic_swap(self, path: Path) -> None:
        """Prepare all the files for a model run in user's directory."""
        self.validate()
        builder = ModelBuilder(model=self, tempdir=path)

        builder.write_inputs()

        logger.info(f"Model files written to {path}")

    def run(
        self, path: str | Path | None = None, silence_warnings: bool = False
    ) -> Result:
        """Run the model using ModelRunner."""
        self.validate()
        path = Path.cwd() if path is None else path
        return ModelRunner(self).run(path, silence_warnings)


def _run_model_with_params(args) -> Result:
    """Helper function to run a model with parameters."""
    model, path, silence_warnings = args
    return model.run(path=path, silence_warnings=silence_warnings)


def run_parallel(
    mls: list[Model],
    path: Path | str | None = None,
    silence_warnings: bool = False,
    **kwargs,
) -> list[Result]:
    """Run multiple models in parallel.

    Parameters:
        mls (list[Model]): List of models to run.
        path (Path | str): The path to the temporary directory.
        silence_warnings (bool): If True, warnings are not raised.
        **kwargs (dict): Keyword arguments for Pool().

    Returns:
        list[Result]: List of results from the model runs.
    """
    with Pool(**kwargs) as pool:
        results = pool.map(
            _run_model_with_params, [(model, path, silence_warnings) for model in mls]
        )

    return results



"""Classes that wrap the crop parameters database for WOFOST (A. de Wit).

From the classes here, only the WOFOSTCropDB is directly accessed by the user,
however, the final usable object will be the CropVariety.

Classes:
    WOFOSTCropFile: Manage a single WOFOST crop file content.
    CropVariety: Manage crop variety parameters.
    WOFOSTCropDB: Manage a single WOFOST crop file content.
"""

from pathlib import Path
from pprint import pprint

from pydantic import BaseModel, computed_field

from pyswap.core.io.io_yaml import load_yaml
from pyswap.libs import crop_params


class WOFOSTCropFile(BaseModel):
    """Manage a single WOFOST crop file content.

    Attributes:
        yaml_content: The entire content of the YAML file.

    Properties:
        metadata: Metadata of the crop file.
        ecotypes: List of eco-types.
        genericc3: Generic settings for C3 crop types.
        genericc4: Generic settings for C4 crop types.
        varieties: List of available varieties.

    Methods:
        get_variety: Get the parameters of a specific variety.
    """

    yaml_content: dict

    @computed_field(return_type=dict)
    def metadata(self):
        """Metadata of the yaml crop file"""
        return self.yaml_content["Metadata"]

    @computed_field(return_type=dict)
    def ecotypes(self):
        return list(self.yaml_content["CropParameters"]["EcoTypes"])

    @computed_field(return_type=dict)
    def genericc3(self):
        """Get generic settings for C3 crop types - plants that bind CO2 into
        3-phosphoglycerate having three carbon atoms. E.g., wheat, rice"""
        return self.yaml_content["CropParameters"]["GenericC3"]

    @computed_field(return_type=dict)
    def genericc4(self):
        """Get generic settings for C4 crop types - plants that bind CO2 into
        oxaloacetate having four carbon atoms. E.g., maize, sugarcane"""
        return self.yaml_content["CropParameters"]["GenericC4"]

    @computed_field(return_type=dict)
    def varieties(self):
        return list(self.yaml_content["CropParameters"]["Varieties"])

    def get_variety(self, variety: str):
        return CropVariety(
            variety=self.yaml_content["CropParameters"]["Varieties"][variety]
        )


class CropVariety(BaseModel):
    """Manage crop variety parameters.

    Attributes:
        variety: Parameters for crop variety from the YAML file (with metadata).

    Properties:
        parameters: Bare parameters of the variety (all metadata removed).
        metadata: The metadata of the variety.
    """

    variety: dict

    @computed_field(return_type=dict)
    def parameters(self):
        params = {
            k.lower(): v[0]
            for k, v in self.variety.items()
            if k != "Metadata" and v[0] != -99.0
        }
        return self._format_tables(params)

    @computed_field(return_type=dict)
    def metadata(self):
        return self.variety["Metadata"]

    @staticmethod
    def _format_tables(table: dict) -> dict[str, list[list]]:
        """pre-format tables from YAML to a list of lists.

        In the YAML file, the tables are seem to be formatted
        in a way where the odd elements in the lists are one
        column and the even elements are the other. This method
        converts this format to a dictionary with two lists.
        """

        formatted = {
            k: [list(row) for row in zip(v[::2], v[1::2], strict=False)]
            if isinstance(v, list)
            else v
            for k, v in table.items()
        }

        return formatted


class WOFOSTCropDB(BaseModel):
    """Simple class for managing crop parameters files.

    Initially, it's meant to be used with A. de Wit's WOFOST crop parameters
    database which is a collection of YAML files. However, it can be easily
    extended to support other formats and databases when they emerge. All
    methods should return the content as WOFOSTCropFile instances. This way
    we ensure that whatever the source format is, the content is always
    usable in the same way.

    Attributes:
        libdir: Path to the directory with crop parameters

    Properties:
        croptypes: List all available crop types (files in the directory)

    Methods:
        load_crop_file: Load a specific crop file and return the content as a
            WOFOSTCropFile instance
    """

    libdir: Path = crop_params

    @computed_field(return_type=None)
    def croptypes(self):
        """Print the list of available files"""
        pprint(load_yaml(crop_params / "crops.yaml")["available_crops"])

    def load_crop_file(self, crop: str) -> WOFOSTCropFile:
        """Load a specific crop file and return the content as a dictionary"""
        path = (
            self.libdir / f"{crop}"
            if crop.endswith(".yaml")
            else self.libdir / f"{crop}.yaml"
        )
        return WOFOSTCropFile(yaml_content=load_yaml(path))


__all__ = ["WOFOSTCropDB"]



import json
import shutil
from pathlib import Path

import pytest
from typer.testing import CliRunner

# Replace with the actual name of your script file
from pyswap.core.cli.cli import app

runner = CliRunner()


@pytest.fixture(scope="function")
def setup_and_teardown():
    # Setup: Define the folder and ensure it's clean
    folder_name = "TestProjectFolder"
    folder_path = Path.cwd() / folder_name

    # Clean up any existing folder
    if folder_path.exists():
        shutil.rmtree(folder_path)

    yield folder_path

    # Teardown: Clean up after the test
    if folder_path.exists():
        shutil.rmtree(folder_path)


def test_cli_init_script(setup_and_teardown):
    folder_path = setup_and_teardown

    inputs = "\n".join(
        [
            "TestProject",
            "1.0",
            "John Doe",
            "XYZ University",
            "john.doe@example.com",
            "",  # Empty comment (user just presses enter)
            "TestProjectFolder",
        ]
    )

    result = runner.invoke(app, ["init", "--script", "--no-pixi"], input=inputs)
    assert result.exit_code == 0

    main_script_path = folder_path / "models" / "main.py"
    assert main_script_path.exists(), "main.py was not created."

    with open(main_script_path) as f:
        content = f.read()
        assert (
            "metadata = psp.components.Metadata(" in content
        ), "Content of main.py is incorrect."


def test_cli_init_notebook(setup_and_teardown):
    folder_path = setup_and_teardown

    inputs = "\n".join(
        [
            "TestProject",
            "1.0",
            "John Doe",
            "XYZ University",
            "john.doe@example.com",
            "",  # Empty comment (user just presses enter)
            "TestProjectFolder",
        ]
    )

    result = runner.invoke(app, ["init", "--notebook", "--no-pixi"], input=inputs)
    assert result.exit_code == 0

    notebook_path = folder_path / "models" / "main.ipynb"
    assert notebook_path.exists(), "main.ipynb was not created."

    with open(notebook_path) as f:
        notebook_content = json.load(f)

        code_cell_source = notebook_content["cells"][2]["source"]
        full_code_content = "".join(code_cell_source)
        assert (
            "metadata = ps.Metadata(" in full_code_content
        ), "Content of main.ipynb is incorrect."


def test_git_initialization_and_gitignore(setup_and_teardown):
    folder_path = setup_and_teardown

    inputs = "\n".join(
        [
            "TestProject",
            "1.0",
            "John Doe",
            "XYZ University",
            "john.doe@example.com",
            "",  # Empty comment (user just presses enter)
            "TestProjectFolder",
        ]
    )

    result = runner.invoke(app, ["init", "--script", "--notebook"], input=inputs)
    assert result.exit_code == 0

    git_dir_path = folder_path / ".git"
    assert (
        git_dir_path.exists() and git_dir_path.is_dir()
    ), ".git directory was not created, Git repo not initialized."

    gitignore_path = folder_path / ".gitignore"
    assert gitignore_path.exists(), ".gitignore file was not created."

    with open(gitignore_path) as f:
        gitignore_content = f.read()
        expected_patterns = [
            "__pycache__/",
            "*.pyc",
            "*.pyo",
            "*.pyd",
            ".env",
            "venv/",
            "env/",
            "*.env",
            "*.ipynb_checkpoints/",
            "data/",
            ".pixi/",
            "pixi.lock",
        ]

        for pattern in expected_patterns:
            assert (
                pattern in gitignore_content
            ), f"Expected pattern '{pattern}' not found in .gitignore file."


def test_pixi_default_behavior(setup_and_teardown):
    """Test that pixi is enabled by default and creates appropriate files."""
    folder_path = setup_and_teardown

    inputs = "\n".join(
        [
            "TestProject",
            "1.0",
            "John Doe",
            "XYZ University",
            "john.doe@example.com",
            "",  # Empty comment (user just presses enter)
            "TestProjectFolder",
        ]
    )

    result = runner.invoke(app, ["init", "--notebook"], input=inputs)
    assert result.exit_code == 0

    # Check that pixi.toml was created
    pixi_toml_path = folder_path / "pixi.toml"
    assert pixi_toml_path.exists(), "pixi.toml was not created by default."

    # Check that README.md was created with pixi instructions
    readme_path = folder_path / "README.md"
    assert readme_path.exists(), "README.md was not created."

    with open(readme_path) as f:
        readme_content = f.read()
        assert "## Environment Setup with Pixi" in readme_content
        assert "pixi install" in readme_content
        assert "pixi run jupyter" in readme_content

    # Check pixi.toml content
    with open(pixi_toml_path) as f:
        pixi_content = f.read()
        assert "TestProject" in pixi_content
        assert "John Doe" in pixi_content
        assert "john.doe@example.com" in pixi_content
        assert "pyswap = \"*\"" in pixi_content


def test_no_pixi_option(setup_and_teardown):
    """Test that --no-pixi disables pixi and creates basic README."""
    folder_path = setup_and_teardown

    inputs = "\n".join(
        [
            "TestProject",
            "1.0",
            "John Doe",
            "XYZ University",
            "john.doe@example.com",
            "",  # Empty comment (user just presses enter)
            "TestProjectFolder",
        ]
    )

    result = runner.invoke(app, ["init", "--notebook", "--no-pixi"], input=inputs)
    assert result.exit_code == 0

    # Check that pixi.toml was NOT created
    pixi_toml_path = folder_path / "pixi.toml"
    assert not pixi_toml_path.exists(), "pixi.toml should not be created with --no-pixi."

    # Check that basic README was created (not README.md)
    readme_path = folder_path / "README"
    assert readme_path.exists(), "Basic README was not created."

    readme_md_path = folder_path / "README.md"
    assert not readme_md_path.exists(), "README.md should not exist with --no-pixi."

    # Check .gitignore doesn't have pixi entries
    gitignore_path = folder_path / ".gitignore"
    with open(gitignore_path) as f:
        gitignore_content = f.read()
        assert ".pixi/" not in gitignore_content
        assert "pixi.lock" not in gitignore_content


def test_upload_swap_success(tmp_path, monkeypatch):
    """Uploading a local executable should copy it to the package location and write version info."""
    import pyswap.utils.executables as exe_mod
    from pyswap.core.cli.cli import app

    runner = CliRunner()

    # Create a fake source executable
    src = tmp_path / "my_swap"
    src.write_text("#!/bin/sh\necho SWAP")
    src.chmod(0o755)

    # Prepare a fake install location and monkeypatch the getter
    install_dir = tmp_path / "install"
    install_dir.mkdir()
    target_path = install_dir / ("swap.exe" if exe_mod.IS_WINDOWS else "swap420")

    monkeypatch.setattr(exe_mod, "get_swap_executable_path", lambda: target_path)

    result = runner.invoke(app, ["upload-swap", str(src), "4.2.0", "--force"])
    assert result.exit_code == 0, result.stdout + str(result.exception)

    # Check file copied
    assert target_path.exists(), "Uploaded executable was not copied to target location"

    # Check version file
    version_file = target_path.parent / "version.yaml"
    assert version_file.exists(), "version.yaml was not written"
    version_txt = version_file.read_text()
    assert "4.2.0" in version_txt


def test_upload_swap_source_missing(tmp_path, monkeypatch):
    """Uploading a non-existent source should return an error."""
    import pyswap.utils.executables as exe_mod
    from pyswap.core.cli.cli import app

    runner = CliRunner()

    missing = tmp_path / "does_not_exist"

    install_dir = tmp_path / "install2"
    install_dir.mkdir()
    target_path = install_dir / ("swap.exe" if exe_mod.IS_WINDOWS else "swap420")
    monkeypatch.setattr(exe_mod, "get_swap_executable_path", lambda: target_path)

    result = runner.invoke(app, ["upload-swap", str(missing), "custom"])
    assert result.exit_code != 0
    assert "Source file not found" in result.stderr or "Error" in result.stderr


def test_upload_swap_target_exists_without_force(tmp_path, monkeypatch):
    """If target exists and --force not provided, upload should fail."""
    import pyswap.utils.executables as exe_mod
    from pyswap.core.cli.cli import app

    runner = CliRunner()

    src = tmp_path / "my_swap2"
    src.write_text("#!/bin/sh\necho SWAP")
    src.chmod(0o755)

    install_dir = tmp_path / "install3"
    install_dir.mkdir()
    target_path = install_dir / ("swap.exe" if exe_mod.IS_WINDOWS else "swap420")
    # create existing target
    target_path.write_text("existing")
    target_path.chmod(0o755)

    monkeypatch.setattr(exe_mod, "get_swap_executable_path", lambda: target_path)

    result = runner.invoke(app, ["upload-swap", str(src), "custom"])
    assert result.exit_code != 0
    assert (
        "already exists" in result.stderr
        or "Use force=True" in result.stderr
        or "Use --force" in result.stderr
    )



"""
Testing module.

!!! note

    Testing in pyswap is currently done through running the test case models
    created with pyswap and comparing the results to those
    obtained from runnign the model the traditional way (using the testcase
    library provided by the developers of SWAP). Currently there is still
    some issues with floating point numbers being properly represented. While
    this is normally not posing significant issue, it should be solved, because
    with more iterations the error resulting from floating point
    misrepresentation will increase. *To avoid throwing errors at testing
    related to this, a tolerance is implemented when comparing the dataframes
    to max 10e-2 on absolute and relative error*.

"""

import pandas as pd

from pyswap import testcase


def test_hupselbrook_model():
    model = testcase.get("hupselbrook")
    result = model.run("./", silence_warnings=True)

    resampled_output = result.csv.resample("YE").sum()

    data = {
        "RAIN": [84.18, 71.98, 80.55],
        "IRRIG": [0.5, 0.0, 0.0],
        "INTERC": [3.74188, 2.05788, 4.91540],
        "RUNOFF": [0.0, 0.0, 0.0],
        "EPOT": [33.10679, 35.99241, 29.89176],
        "EACT": [16.68715, 17.17961, 17.88879],
        "DRAINAGE": [22.11357, 26.44815, 24.76249],
        "QBOTTOM": [0.0, 0.0, 0.0],
        "GWL": [-1107.65824, -1154.37603, -1036.83171],
        "TPOT": [38.71198, 29.41787, 32.57304],
        "TACT": [38.17328, 29.21504, 32.57304],
        "DSTOR": [3.96418, -2.92064, 0.41029],
    }

    index = pd.to_datetime(["2002-12-31", "2003-12-31", "2004-12-31"])

    expected_data = pd.DataFrame(data, index=index)
    expected_data.index.name = "DATETIME"
    expected_data = expected_data.resample("YE").sum()

    # Compare the result with the expected values
    pd.testing.assert_frame_equal(
        resampled_output,
        expected_data,
        check_dtype=False,
        rtol=1e-2,
    )


def test_grassgrowth():
    model = testcase.get("grassgrowth")
    result = model.run("./", silence_warnings=True)

    expected_data = pd.DataFrame(
        {
            "PGRASSDM": {
                pd.Timestamp("1980-12-31 00:00:00"): 1375.2639006557376,
                pd.Timestamp("1981-12-31 00:00:00"): 1396.1866007397261,
                pd.Timestamp("1982-12-31 00:00:00"): 1907.5665839178082,
                pd.Timestamp("1983-12-31 00:00:00"): 1958.4105762739725,
                pd.Timestamp("1984-12-31 00:00:00"): 1506.5268562841532,
            },
            "GRASSDM": {
                pd.Timestamp("1980-12-31 00:00:00"): 1289.721095819672,
                pd.Timestamp("1981-12-31 00:00:00"): 1318.114862739726,
                pd.Timestamp("1982-12-31 00:00:00"): 1752.5114093972602,
                pd.Timestamp("1983-12-31 00:00:00"): 1786.5715040821917,
                pd.Timestamp("1984-12-31 00:00:00"): 1397.4912103005465,
            },
            "PMOWDM": {
                pd.Timestamp("1980-12-31 00:00:00"): 8582.550397349727,
                pd.Timestamp("1981-12-31 00:00:00"): 8498.116628602738,
                pd.Timestamp("1982-12-31 00:00:00"): 9036.348089232877,
                pd.Timestamp("1983-12-31 00:00:00"): 9446.297125178082,
                pd.Timestamp("1984-12-31 00:00:00"): 7937.071480737705,
            },
            "MOWDM": {
                pd.Timestamp("1980-12-31 00:00:00"): 7415.0520341530055,
                pd.Timestamp("1981-12-31 00:00:00"): 7387.342750109589,
                pd.Timestamp("1982-12-31 00:00:00"): 7713.840507123287,
                pd.Timestamp("1983-12-31 00:00:00"): 8111.034105342465,
                pd.Timestamp("1984-12-31 00:00:00"): 6818.060586885246,
            },
        }
    )

    expected_data.index.name = "DATETIME"
    expected_data_resampled = expected_data.resample("YE").mean()

    resampled_output = result.csv.resample("YE").mean()

    pd.testing.assert_frame_equal(
        resampled_output,
        expected_data_resampled,
        check_dtype=False,
        rtol=1e-2,
    )


def test_simple_model():
    model = testcase.get("simple_test_model")
    model.run("./", silence_warnings=True)

    # TODO add expected data



"""GIS functionality for pyswap.

Currently this module provides limited functionality with respect to GIS data.
However, it is intended to be expanded in the future.

Modules:
    location: Location class and related functionality.
"""

from pyswap.gis.location import Location

__all__ = ["Location"]



import pandas as pd

import pyswap.db.co2concentration as co2


def test_get_filtered_co2file():
    co2_db = co2.CO2concentration()

    # Test with a valid period
    period = ["2000-01-01", "2020-12-31"]
    data = co2_db.get_filtered_co2file(period)

    # Load expected data
    expected_data = pd.read_csv(
        "tests/test_co2concentration/Atmospheric.co2",
        delimiter=" ",
        index_col=0,
    )

    # Check if the data is equal
    pd.testing.assert_frame_equal(
        expected_data,
        data,
        check_dtype=False,
    )


def test_write_co2():
    co2_db = co2.CO2concentration()

    # Test with a valid period
    period = ["2000-01-01", "2020-12-31"]
    directory = "tests/test_co2concentration/write/"
    co2_db.write_co2(directory, period)


if __name__ == "__main__":
    test_get_filtered_co2file()
    # test_write_co2()



"""Variables shared across the package.

Variables:
    IS_WINDOWS: Whether the system is Windows.
    BASE_PATH: Representation of file path depending on the system.
    root: Root directory of the package.
    validation_rules: Path to YAML validation rules.
    VALIDATIONRULES: Validation rules for the package as dictionary.
"""

import platform
from importlib import resources
from pathlib import Path

from pyswap.core.io.io_yaml import load_yaml

IS_WINDOWS = platform.system() == "Windows"
"""Whether the system is Windows."""

BASE_PATH = ".\\" if IS_WINDOWS else "./"
"""Representation of file path depending on the system."""

root: Path = resources.files("pyswap")
"""Root directory of the package."""

validation_rules: Path = root / "core" / "validation.yaml"
VALIDATIONRULES = load_yaml(validation_rules)

FNAME_IN: str = "swap"
"""Generic name for the SWAP input files."""

FNAME_OUT: str = "result"
"""Generic name for the SWAP output files."""

ADDITIONAL_SWITCHES: list[str] = [
    "schedule",
    "isuas",
    "tcs",
    "tcsfix",
    "dcs",
    "dcslim",
    "dramet",
    "ipos",
    "idev",
    "idsl",
]

EXTENSIONS: list[str] = [
    "wba",
    "end",
    "vap",
    "bal",
    "blc",
    "sba",
    "ate",
    "bma",
    "drf",
    "swb",
    "ini",
    "inc",
    "crp",
    "str",
    "irg",
    "csv",
    "csv_tz",
]

EXTENSION_SWITCHES: list[str] = [f"sw{ext}" for ext in EXTENSIONS]



"""Plotting functionality for pySWAP.

Plotting of results depends heavily on the application of the model. Therefore,
in this module, I only inclduded a few examples of how to plot results for
specific cases.

Modules:
    evapotranspiration: Functions for plotting evapotranspiration data.
    watercontent: Functions for plotting water content data as heat map.
"""

from pyswap.core.plot.evapotranspiration import evapotranspiration
from pyswap.core.plot.watercontent import water_content

__all__ = ["evapotranspiration", "water_content"]



"""Module processing ASCII files in SWAP format.

Steps to process SWAP format ASCII files:
1. Identify and remove comments.
2. Parse the remaining content into key-value pairs, tables and arrays.
"""

import logging
import re

from pyswap.core.io.processors import TableProcessor

logger = logging.getLogger(__name__)


def remove_comments(text: str) -> str:
    """Remove comments from a SWAP input file.

    In a SWAP input files some lines are comments. Full line comments
    start with a * character. Partial comments start with a ! character and
    sometimes follow the actual data.

    !!! note
        Empty lines are not removed at this stage. They are important for
        parsing the tables.

    Parameters:
        text (str): The text to remove comments from.

    Returns:
        str: Stripped text with comments removed.
    """

    text = re.sub(r"(^\*.*$|!.*)", "", text, flags=re.MULTILINE)

    return text.strip()


def parse_ascii_file(file_content, grass=False) -> dict[str, dict]:
    """Parse an ASCII file in SWAP format.

    !!! note "Assumptions"
        - key-value pairs are lines with a single `=` character
        - tables are lines in which columns are split by spaces
        - empty tags are lines that end with an `=` character, followed by
            table-like data in the following lines.
        - tables are followed by an empty line or a line that is not
          a part of another table.

    Parameters:
        file_content (str): The content of the ASCII file.

    Returns:
        dict: A dictionary with key-value pairs, arrays and tables
            (in the exact order).
    """
    cleaned_file = remove_comments(file_content)
    lines = cleaned_file.splitlines()
    pairs = {}
    arrays = {}
    tables = {}

    def is_key_value(line):
        return (
            "=" in line
            and not line.strip().startswith("=")
            and not line.strip().endswith("=")
        )

    def format_key_value(line):
        key, value = line.split("=", 1)
        return {key.strip().lower(): value.strip()}

    def is_table(line):
        """Check if the line is a part of a table.

        A table is essentially everything else than a key-value pair or
        an empty tag except empty lines.
        """
        return line.strip() and "=" not in line and not line.strip().endswith("=")

    def is_empty_tag(line):
        """Check if the line is an empty tag.

        An empty tag is a line where there is only the tag folloed by an = sign (e.g., DZNEW =)
        and the data for that tag is in the next line(s). This is most common for tables,
        which in pySWAP are called ARRAYS - tables with no header, but values groupped in
        columns separated by spaces."""

        return line.strip().endswith("=")

    def parse_table(lines, start_index, key, param_type):
        """Parse a table from the list of lines.

        This function is triggered if a line is detected as an empty tag or a table. It will
        assume all lines after the empty tag or the table header are part of the table until
        an empty line or a line that is not part of the table is found. Those lines are then
        stored in a list, later used to skip the table rows before parsing the next item.
        """
        data = []
        tp = TableProcessor()
        for line in lines[start_index:]:
            if line.strip() and not is_key_value(line) and not is_empty_tag(line):
                data.append(line.strip().split())
            else:
                break
        processed = tp.process(
            data_type=param_type,
            data=data,
            columns=tuple(key.strip().split()),
            grass=grass,
        )
        logger.debug(f"Processed {len(data)} rows for {key.strip()}")
        return processed, len(data)

    i = 0
    # loop over the list of lines, stripping each

    while i < len(lines):
        line = lines[i].strip()

        if is_key_value(line):
            pairs.update(format_key_value(line))

        elif is_empty_tag(line):
            key = line[:-1].strip()
            array = parse_table(
                lines=lines, start_index=i + 1, key=key, param_type="array"
            )
            arrays.update(array[0])
            i += array[1] + 1  # Skip the tag data

        elif is_table(line):
            # The table header is the line itself (de facto dictionary key)
            table = parse_table(
                lines=lines, start_index=i + 1, key=line, param_type="table"
            )
            tables.update(table[0])
            i += table[1] + 1  # Skip the table rows
        i += 1  # Move to the next line

    return pairs | arrays | tables



# mypy: disable-error-code="call-overload, misc, operator"
# the operator error was raised bacause of Pydantic validators. The current
# implementation is correct, so it was not the priority to fix.


"""Meteorology settings and data.

This module contains the classes and functions to handle meteorological settings and data for simulations.

Classes:
    MetFile: Meteorological data for the .met file.
    Meteorology: Meteorological settings of the simulation.

Functions:
    load_from_csv: Load meteorological data from a CSV file.
    load_from_knmi: Load meteorological data from KNMI API.
"""

from datetime import datetime as _datetime
from typing import (
    Literal as _Literal,
)

from knmi import (
    get_day_data_dataframe as _get_day_data_dataframe,
    get_hour_data_dataframe as _get_hour_data_dataframe,
)
from numpy import exp as _exp
from pydantic import (
    Field as _Field,
    PrivateAttr as _PrivateAttr,
)

from pyswap.components.tables import (
    DAILYMETEODATA,
    DETAILEDRAINFALL,
    RAINFLUX,
    SHORTINTERVALMETEODATA,
)
from pyswap.core.basemodel import PySWAPBaseModel as _PySWAPBaseModel
from pyswap.core.fields import (
    Decimal2f as _Decimal2f,
    File as _File,
    String as _String,
    Table as _Table,
)
from pyswap.core.io.io_csv import load_csv
from pyswap.core.valueranges import UNITRANGE as _UNITRANGE
from pyswap.gis import Location as _Location
from pyswap.utils.mixins import (
    FileMixin as _FileMixin,
    SerializableMixin as _SerializableMixin,
    YAMLValidatorMixin as _YAMLValidatorMixin,
)

__all__ = [
    "DAILYMETEODATA",
    "DETAILEDRAINFALL",
    "RAINFLUX",
    "SHORTINTERVALMETEODATA",
    "MetFile",
    "Meteorology",
    "metfile_from_csv",
    "metfile_from_knmi",
]


class MetFile(_PySWAPBaseModel, _FileMixin, _SerializableMixin):
    """Meteorological data for the .met file.

    This object is created by functions fetching or loading meteorological data
    from various sources. The data is stored as a pandas.DataFrame, but
    is formatted with a custom field serializer of the CSVTable field type.

    Attributes:
        metfil (str): name of the .met file
        content (CSVTable): meteorological data file
    """

    # None, because the extension has to be added to metfil
    _extension: bool = _PrivateAttr(default=None)

    metfil: _String
    content: _Table | None = _Field(default=None, exclude=True)


class Meteorology(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Meteorological settings of the simulation.

    !!! note
        SWRAIN and SWETSINE should be optional,
        but Fortran code evaluates its presence anyway. They are set to
        0 by default.


    Attributes:
        meteo_location (Location): a point GIS object. If provided, lat
            and alt must not be provided. By default they are overwritten.
        lat (Decimal): latitude of the meteo station [degrees].
        swetr (int): Switch type of weather data for
            potential evapotranspiration:

            * 0 - Use basic weather data and apply Penman-Monteith equation.
            * 1 - Use reference evapotranspiration data in combination with
                crop factors.

        swdivide (int): Switch for distribution of E and T. Defaults to 0:

            * 0 - Based on crop and soil factors.
            * 1 - Based on direct application of Penman-Monteith.

        swmetdetail (int): Switch for time interval of evapotranspiration and
            rainfall weather data:

            * 0 - Daily data.
            * 1 - Subdaily data.

        swrain (int): Switch for use of actual rainfall intensity,
            defaults to 0:

            * 0 - Use daily rainfall amounts.
            * 1 - Use daily rainfall amounts + mean intensity.
            * 2 - Use daily rainfall amounts + duration.
            * 3 - Use detailed rainfall records (dt < 1 day), as supplied in
                separate file.

        swetsine (int): Switch, distribute daily Tp and Ep according to
            sinus wave, default to 0:

            * 0 - No distribution.
            * 1 - Distribute Tp and Ep according to sinus wave.

        metfile (MetFile): MetFile model containing meteorological data to
            be saved to .met file.
        alt (Decimal): Altitude of the meteo station [m].
        altw (Decimal): Altitude of the wind [m].
        angstroma (Decimal): Fraction of extraterrestrial radiation reaching
            the earth on overcast days.
        angstromb (Decimal): Additional fraction of extraterrestrial radiation
            reaching the earth on clear days.
        table_rainflux (Table): rainfall intensity RAINFLUX as function
            of time TIME.
        rainfil (str): file name of file with detailed rainfall data.
        nmetdetail (int): Number of weather data records each day.

    Properties:
        met: Returns the string representation of the met file.

    Methods:
        write_met: Writes the .met file.
    """

    lat: _Decimal2f | None = _Field(default=None, ge=-90, le=90)
    meteo_location: _Location | None = _Field(default=None, exclude=True)
    swetr: _Literal[0, 1] | None = None
    swdivide: _Literal[0, 1] | None = None
    swrain: _Literal[0, 1, 2, 3] | None = 0
    swetsine: _Literal[0, 1] = 0
    metfile: _File | None = _Field(default=None, repr=False)
    alt: _Decimal2f | None = _Field(default=None, ge=-400.0, le=3000.0)
    altw: _Decimal2f = _Field(default=None, ge=0.0, le=99.0)
    angstroma: _Decimal2f = _Field(default=None, **_UNITRANGE)
    angstromb: _Decimal2f = _Field(default=None, **_UNITRANGE)
    swmetdetail: _Literal[0, 1] | None = None
    table_rainflux: _Table | None = None
    rainfil: _String | None = None
    nmetdetail: int | None = _Field(default=None, ge=1, le=96)

    @property
    def met(self):
        return self.metfile.content.to_csv(index=False, lineterminator="\n")

    def model_post_init(self, __context=None):
        """Set lat, and alt from `meteo_location` if Location object is provided."""
        if self.meteo_location:
            self.lat = self.meteo_location.lat
            self.alt = self.meteo_location.alt

        self._validation = True
        self.validate_with_yaml()
        self._validation = False

    def write_met(self, path: str):
        """Write the .met file.

        !!! note

            in this function the extension is not passed because
            swp file requires the metfile parameter to be passed already with
            the extension.

        Parameters:
            path (str): Path to the file.
        """

        self.metfile.save_file(string=self.met, fname=self.metfile.metfil, path=path)


def metfile_from_csv(metfil: str, csv_path: str, **kwargs) -> MetFile:
    """Method for loading daily meteorological data from a CSV file.

    !!! note

        The CSV file must contain the following columns:
        station, dd, mm, yyyy, tmin, tmax, hum, wet, wind, rain, etref, rad

    Parameters:
        metfil (str): name of the .met file
        csv_path (str): path to the CSV file
        **kwargs (dict): keyword arguments for pandas.read_csv

    Returns:
        MetFile object.
    """
    # Create table from csv
    df = load_csv(csv_path, **kwargs)
    table = DAILYMETEODATA.create(data=df.to_dict())

    # Make sure Station column has quotes
    table.loc[:, "STATION"] = table.STATION.apply(
        lambda x: f"'{x}'" if not str(x).startswith("'") else x
    )

    return MetFile(metfil=metfil, content=table)


def metfile_from_knmi(
    metfil: str,
    stations: str | list,
    start: str | _datetime = "20000101",
    end: str | _datetime = "20200101",
    frequency: _Literal["day", "hour"] = "day",
    # inseason: bool = False,  # Do not use, will cause missing data crashing SWAP
) -> MetFile:
    """Retrieves the meteorological data from KNMI API using knmi-py and
    enforces SWAP required format.

    !!! note:
        Currently, only daily data can be retrieved.

    Parameters:
        metfil (str): name of the .met file
        stations (str | list): station number(s) to retrieve data from
        start (str | dt): start date of the data
        end (str | dt): end date of the data
        frequency (Literal['day', 'hour']): frequency of the data (day or hour)

    Returns:
        MetFile object.
    """

    if not isinstance(stations, list):
        stations = [stations]

    # variables to retrieve and their SWAP name
    variables = {
        "STN": "STATION",
        "Q": "RAD",
        "TN": "TMIN",
        "TX": "TMAX",
        "UG": "HUM",
        "FG": "WIND",
        "RH": "RAIN",
        "EV24": "ETREF",
        "DR": "WET",
    }

    # Retrieve data
    get_func = (
        _get_day_data_dataframe if frequency == "day" else _get_hour_data_dataframe
    )

    df = get_func(
        stations=stations,
        start=start,
        end=end,
        variables=list(variables.keys())[1:],
    )

    # Making separate columns for day, month, year
    df.insert(1, "DD", df.index.day)
    df.insert(2, "MM", df.index.month)
    df.insert(3, "YYYY", df.index.year)

    # Rename columns and drop index
    df = df.reset_index(drop=True)
    df = df.rename(columns=variables)

    # Set -1 to zero in RAIN and SUNH columns (rain or sun hours < 0.05 mm and h respectively)
    df["RAIN"] = df["RAIN"].apply(lambda x: 0 if x == -1 else x)

    # Changing unit of data (see knmi documentation)
    factor = {
        "RAD": 10,  # Convert from J/cm2 to kJ/m2
        "TMIN": 0.1,  # Unit: Convert from 1 to 0.1 degC
        "TMAX": 0.1,  # Convert from 1 to 0.1 degC
        "HUM": 0.01,  # Convert from % to fraction
        "WIND": 0.1,  # Convert from 1 to 0.1 m/s
        "RAIN": 0.1,  # Convert from 1 to 0.1 mm
        "ETREF": 0.1,  # Convert from 1 to 0.1 mm
        "WET": (0.1 / 24),  # Convert from 0.1 h to fraction of day
    }
    df = df.apply(lambda x: x * factor.get(x.name, 1))

    # Convert from fraction to kPa according to Allen et al. (1998)
    es_min = 0.6108 * _exp(17.27 * df["TMIN"].values / (df["TMIN"].values + 237.3))
    es_max = 0.6108 * _exp(17.27 * df["TMAX"].values / (df["TMAX"].values + 237.3))
    df["HUM"] = (es_min + es_max) / 2 * df["HUM"]

    # Make sure Station column has quotes
    df["STATION"] = df["STATION"].apply(
        lambda x: f"'{x}'" if not str(x).startswith("'") else x
    )

    # Make MeteoData table
    table = DAILYMETEODATA.create(data=df.to_dict())

    return MetFile(metfil=metfil, content=table)
